import os
import sys
import matplotlib.pyplot as plt
import numpy as np

import torch
import torch.nn.parallel
from modules.python.models.test_debug import test
from modules.python.models.ModelHander import ModelHandler
from modules.python.TextColor import TextColor
from modules.python.FileManager import FileManager
from modules.python.Options import ImageSizeOptions
"""
FREEZE THIS BRANCH TO HAVE 1 WINDOW!!
Train a model and save the model that performs best.

Input:
- A train CSV containing training image set information (usually chr1-18)
- A test CSV containing testing image set information (usually chr19)

Output:
- A trained model
"""


def save_rle_confusion_matrix(stats_dictionary, output_directory):
    # plot confusion matrix
    fig, ax = plt.subplots(figsize=(20, 20))
    cf = np.array(stats_dictionary['rle_confusion_matrix'], dtype=np.int)
    im = ax.imshow(cf)
    rle_labels = [str(i) for i in range(0, ImageSizeOptions.TOTAL_RLE_LABELS)]

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(rle_labels)))
    ax.set_yticks(np.arange(len(rle_labels)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(rle_labels)
    ax.set_yticklabels(rle_labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    for i in range(len(rle_labels)):
        for j in range(len(rle_labels)):
            if cf[i, j] > 0:
                if i == j:
                    text = ax.text(j, i, cf[i, j], ha="center", va="center", color="g")
                else:
                    text = ax.text(j, i, cf[i, j], ha="center", va="center", color="r")

    ax.set_title("RLE Confusion Matrix")
    fig.tight_layout()
    # plt.show()
    plt.savefig(output_directory + "/RLE_CONFUSION_MATRIX.png", dpi=100)


def save_base_confusion_matrix(stats_dictionary, output_directory):
    # plot confusion matrix
    fig, ax = plt.subplots(figsize=(5, 5))
    cf = np.array(stats_dictionary['base_confusion_matrix'], dtype=np.int)
    im = ax.imshow(cf)
    base_labels = [str(i) for i in ['-', 'A', 'C', 'T', 'G']]

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(base_labels)))
    ax.set_yticks(np.arange(len(base_labels)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(base_labels)
    ax.set_yticklabels(base_labels)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    for i in range(len(base_labels)):
        for j in range(len(base_labels)):
            if cf[i, j] > 0:
                if i == j:
                    text = ax.text(j, i, cf[i, j], ha="center", va="center", color="g")
                else:
                    text = ax.text(j, i, cf[i, j], ha="center", va="center", color="r")

    ax.set_title("BASE Confusion Matrix")
    fig.tight_layout()
    # plt.show()
    plt.savefig(output_directory + "/BASE_CONFUSION_MATRIX.png", dpi=100)


def test_interface(test_file, batch_size, gpu_mode, num_workers, model_path, output_directory, print_details):
    """
    Test a trained model
    :param test_file: Path to directory containing test images
    :param batch_size: Batch size for training
    :param gpu_mode: If true the model will be trained on GPU
    :param num_workers: Number of workers for data loading
    :param model_path: Path to a saved model
    :param output_directory: Path to output_directory
    :param print_details: If true then logs will be printed
    :return:
    """
    sys.stderr.write(TextColor.PURPLE + 'Loading data\n' + TextColor.END)

    output_directory = FileManager.handle_output_directory(output_directory)

    if os.path.isfile(model_path) is False:
        sys.stderr.write(TextColor.RED + "ERROR: INVALID PATH TO MODEL\n")
        exit(1)

    sys.stderr.write(TextColor.GREEN + "INFO: MODEL LOADING\n" + TextColor.END)

    transducer_model, hidden_size, gru_layers, prev_ite = \
        ModelHandler.load_simple_model(model_path,
                                       input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                       image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                       seq_len=ImageSizeOptions.SEQ_LENGTH,
                                       num_base_classes=ImageSizeOptions.TOTAL_BASE_LABELS,
                                       num_rle_classes=ImageSizeOptions.TOTAL_RLE_LABELS)

    sys.stderr.write(TextColor.GREEN + "INFO: MODEL LOADED\n" + TextColor.END)

    if print_details and gpu_mode:
        sys.stderr.write(TextColor.GREEN + "INFO: GPU MODE NOT AVAILABLE WHEN PRINTING DETAILS. "
                                           "SETTING GPU MODE TO FALSE.\n" + TextColor.END)
        gpu_mode = False

    if gpu_mode:
        transducer_model = transducer_model.cuda()

    stats_dictionary = test(test_file, batch_size, gpu_mode, transducer_model, num_workers,
                            gru_layers, hidden_size, num_base_classes=ImageSizeOptions.TOTAL_BASE_LABELS,
                            num_rle_classes=ImageSizeOptions.TOTAL_RLE_LABELS,
                            output_directory=output_directory,
                            print_details=print_details)

    save_rle_confusion_matrix(stats_dictionary, output_directory)
    save_base_confusion_matrix(stats_dictionary, output_directory)

    sys.stderr.write(TextColor.PURPLE + 'DONE\n' + TextColor.END)


import sys
import torch
import torchnet.meter as meter
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
from tqdm import tqdm
from modules.python.models.dataloader_debug import SequenceDataset
from modules.python.TextColor import TextColor
from modules.python.Options import ImageSizeOptions, TrainOptions
"""
WARNING: THIS IS A DEBUGGING TOOL INTENDED TO BE USED BY THE DEVELOPERS ONLY.
"""
# CLASS_WEIGHTS = [0.3, 1.0, 1.0, 1.0, 1.0]


def precision(label, confusion_matrix):
    col = confusion_matrix[:, label]
    return round(confusion_matrix[label, label] / max(1, col.sum()) , 3)


def label_to_literal(label):
    label_decoder = {'A': 0, 'C': 20, 'G': 40, 'T': 60, '_': 0}
    if label == 0:
        return '-', 0
    if label <= 20:
        return 'A', label
    if label <= 40:
        return 'C', label - 20
    if label <= 60:
        return 'G', label - 40

    return 'T', label - 60


label_decoder = {1: 'A', 2: 'C', 3: 'G', 4: 'T', 0: '-'}


def test(data_file, batch_size, gpu_mode, transducer_model, num_workers, gru_layers, hidden_size, output_directory,
         num_base_classes, num_rle_classes, print_details=False):
    # transformations = transforms.Compose([transforms.ToTensor()])
    # torch.set_num_threads(num_workers)
    # data loader
    test_data = SequenceDataset(data_file)
    test_loader = DataLoader(test_data,
                             batch_size=batch_size,
                             shuffle=False,
                             num_workers=num_workers,
                             pin_memory=gpu_mode)
    # sys.stderr.write(TextColor.CYAN + 'Test data loaded\n')

    # set the evaluation mode of the model
    transducer_model.eval()

    test_file_logger = None
    if print_details:
        test_file_logger = open(output_directory + "prediction_debug.txt", 'w')
        sys.stdout = test_file_logger

    # class_weights = torch.Tensor(CLASS_WEIGHTS)
    # Loss not doing class weights for the first pass
    criterion_base = nn.CrossEntropyLoss()
    criterion_rle = nn.CrossEntropyLoss()

    if gpu_mode is True:
        criterion_base = criterion_base.cuda()
        criterion_rle = criterion_rle.cuda()

    # Test the Model
    sys.stderr.write(TextColor.PURPLE + 'Test starting\n' + TextColor.END)
    base_confusion_matrix = meter.ConfusionMeter(num_base_classes)
    rle_confusion_matrix = meter.ConfusionMeter(num_rle_classes)

    total_loss = 0
    total_loss_rle = 0
    total_images = 0
    accuracy = 0
    prediction_name_set = set()
    with torch.no_grad():
        pbar = tqdm(total=len(test_loader), desc='Accuracy: ', leave=True, ncols=100)
        for ii, (images, label_base, label_run_length, position, contig, contig_start, contig_end, chunk_id, filename) in \
                enumerate(test_loader):
            images = images.type(torch.FloatTensor)
            label_base = label_base.type(torch.LongTensor)
            label_rle = label_run_length.type(torch.LongTensor)
            if gpu_mode:
                # encoder_hidden = encoder_hidden.cuda()
                images = images.cuda()
                label_base = label_base.cuda()
                label_rle = label_rle.cuda()
            hidden = torch.zeros(images.size(0), 2 * TrainOptions.GRU_LAYERS, TrainOptions.HIDDEN_SIZE)
            if gpu_mode:
                hidden = hidden.cuda()
            for i in range(0, ImageSizeOptions.SEQ_LENGTH, TrainOptions.WINDOW_JUMP):
                if i + TrainOptions.TRAIN_WINDOW > ImageSizeOptions.SEQ_LENGTH:
                    break
                image_chunk = images[:, i:i+TrainOptions.TRAIN_WINDOW]
                label_base_chunk = label_base[:, i:i+TrainOptions.TRAIN_WINDOW]
                label_rle_chunk = label_rle[:, i:i+TrainOptions.TRAIN_WINDOW]
                position_chunk = position[:, i:i+TrainOptions.TRAIN_WINDOW]
                output_base, output_rle, hidden = transducer_model(image_chunk, hidden)

                # do softmax and get prediction
                m = nn.Softmax(dim=2)
                soft_probs = m(output_base)
                output_preds = soft_probs.cpu()
                base_max_value, predicted_base_label = torch.max(output_preds, dim=2)

                # convert everything to list
                predicted_base_label = predicted_base_label.numpy().tolist()

                # do softmax and get prediction for rle
                m_rle = nn.Softmax(dim=2)
                rle_soft_probs = m_rle(output_rle)
                rle_output_preds = rle_soft_probs.cpu()
                rle_max_value, predicted_rle_labels = torch.max(rle_output_preds, dim=2)
                predicted_rle_labels = predicted_rle_labels.data.numpy().tolist()

                true_bases = label_base_chunk.cpu().contiguous().numpy().tolist()
                true_rles = label_rle_chunk.cpu().numpy().tolist()

                for i in range(image_chunk.size(0)):
                    column_count = 0
                    for pred_rle, true_rle, pred_base, true_base, pos in zip(predicted_rle_labels[i], true_rles[i],
                                                                             predicted_base_label[i], true_bases[i],
                                                                             position_chunk[i]):
                        if pred_rle != true_rle[0]:
                            prediction_name = str(contig[i]) + "_" + str(contig_start[i].item()) + "_" \
                                              + str(contig_end[i].item()) + "_" + str(chunk_id[i].item()) + "_" \
                                              + str(pos[0].item()) + "_" + str(pos[1].item()) + "_" + str(pos[2].item())
                            if prediction_name in prediction_name_set:
                                column_count += 1
                                continue
                            prediction_name_set.add(prediction_name)

                            if print_details:
                                print("RLE TRUE/PRED: " + "{:02d}/{:02d}".format(true_rle[0] , pred_rle) + ",",
                                      "BASE TRUE/PRED: " + "{}/{}".format(label_decoder[true_base[0]], label_decoder[pred_base])+ ",",
                                      "CONTIG: " + (str(contig[i]) + ":" + str(contig_start[i].item()) + "-" + str(contig_end[i].item()))+ ",",
                                      "CHUNK ID: " + str(chunk_id[i].item())+ ",",
                                      "POS: {:3d} {:3d} INDEX: {:2d} SPLIT INDEX:{:2d}".format(pos[0].item() + contig_start[i].item(), pos[0].item(), pos[1].item(), pos[2].item())+ ",",
                                      "FILENAME: ", str(filename[i])+ ",")
                                print(''.join(['-'] * 93))
                                print("R: ", ' '.join(["{:5d}".format(int(x/2)) if x % 2 == 0 else ' ' for x in range(0, 21)]), '  |')
                                print(''.join(['-'] * 93))
                                print("A: ", ' '.join(["{:3d}".format(int(x)) for x in image_chunk[i, column_count].numpy().tolist()[0:22]]), '|')
                                print("C: ", ' '.join(["{:3d}".format(int(x)) for x in image_chunk[i, column_count].numpy().tolist()[22:44]]), '|')
                                print("G: ", ' '.join(["{:3d}".format(int(x)) for x in image_chunk[i, column_count].numpy().tolist()[44:66]]), '|')
                                print("T: ", ' '.join(["{:3d}".format(int(x)) for x in image_chunk[i, column_count].numpy().tolist()[66:88]]), '|')
                                print("*: ", ' '.join(["{:3d}".format(int(x)) for x in image_chunk[i, column_count].numpy().tolist()[88:]]), ' '*79, '|')
                                print(''.join(['-'] * 93))

                        column_count += 1

                loss_base = criterion_base(output_base.contiguous().view(-1, num_base_classes),
                                           label_base_chunk.contiguous().view(-1))
                loss_rle = criterion_rle(output_rle.contiguous().view(-1, num_rle_classes),
                                         label_rle_chunk.contiguous().view(-1))
                loss = loss_base + loss_rle
                base_confusion_matrix.add(output_base.data.contiguous().view(-1, num_base_classes),
                                          label_base_chunk.data.contiguous().view(-1))
                rle_confusion_matrix.add(output_rle.data.contiguous().view(-1, num_rle_classes),
                                         label_rle_chunk.data.contiguous().view(-1))
                total_loss += loss.item()
                total_images += images.size(0)
                total_loss_rle += loss_rle.item()

            base_cm_value = base_confusion_matrix.value()
            rle_cm_value = rle_confusion_matrix.value()

            base_denom = base_cm_value.sum()
            # rle_denom = rle_cm_value.sum() - rle_cm_value[0][0]
            rle_denom = rle_cm_value.sum()
            base_corrects = 0

            for label in range(0, ImageSizeOptions.TOTAL_BASE_LABELS):
                base_corrects = base_corrects + base_cm_value[label][label]
            rle_corrects = 0

            for label in range(0, ImageSizeOptions.TOTAL_RLE_LABELS):
                rle_corrects = rle_corrects + rle_cm_value[label][label]
                # rle_denom = rle_denom - rle_cm_value[0][label]

            # calculate the accuracy
            base_accuracy = 100.0 * (base_corrects / max(1.0, base_denom))
            rle_accuracy = 100.0 * (rle_corrects / max(1.0, rle_denom))

            # set the tqdm bar's accuracy and loss value
            pbar.update(1)
            pbar.set_description("Base acc: " + str(round(base_accuracy, 4)) +
                                 ", RLE acc: " + str(round(rle_accuracy, 4)))

    avg_loss = total_loss / total_images if total_images else 0
    np.set_printoptions(threshold=np.inf)
    pbar.close()

    sys.stderr.write(TextColor.YELLOW+'\nTest Loss: ' + str(avg_loss) + "\n"+TextColor.END)
    sys.stderr.write(TextColor.BLUE + "Base Confusion Matrix: \n" + str(base_confusion_matrix.value()) + "\n" + TextColor.END)
    sys.stderr.write(TextColor.RED + "RLE Confusion Matrix: \n" + str(rle_confusion_matrix.value()) + "\n" + TextColor.END)
    # sys.stderr.write("label\t\tprecision\n")
    # for label in range(0, ImageSizeOptions.TOTAL_LABELS):
    #     sys.stderr.write(str(label_to_literal(label)) + '\t' + str(precision(label, confusion_matrix.conf)) + "\n")

    return {'loss': avg_loss, 'accuracy': accuracy, 'base_confusion_matrix': base_confusion_matrix.conf,
            'rle_confusion_matrix': rle_confusion_matrix.conf}



import sys
import torch
from tqdm import tqdm
import torchnet.meter as meter
import torch.nn as nn
from torch.utils.data import DataLoader
import numpy as np
from modules.python.models.dataloader import SequenceDataset
from modules.python.TextColor import TextColor
from modules.python.Options import ImageSizeOptions, TrainOptions
"""
This script implements the test method.
The test method evaluates a trained model on a given test dataset and report various accuracy matrices.
"""


def test(data_filepath,
         batch_size,
         gpu_mode,
         transducer_model,
         num_workers,
         gru_layers,
         hidden_size,
         num_base_classes,
         num_rle_classes,
         print_details=False):
    """
    This method performs testing of a trained model.
    :param data_filepath: Path to a directory containing all labeled h5 files used for testing
    :param batch_size: Size of minibatch
    :param gpu_mode: If true, the model will be loaded on GPU using CUDA.
    :param transducer_model: A trained model
    :param num_workers: Number of thread workers for dataloader
    :param gru_layers: Number of GRU layers
    :param hidden_size: Hidden size of the model.
    :param num_base_classes: Number of classes for base prediction
    :param num_rle_classes: Number of classes for RLE prediction
    :param print_details: A debug parameter.
    :return:
    """
    # data loader
    test_data = SequenceDataset(data_filepath)
    test_loader = DataLoader(test_data,
                             batch_size=batch_size,
                             shuffle=False,
                             num_workers=num_workers,
                             pin_memory=gpu_mode)

    # set the model to evaluation mode
    transducer_model.eval()

    # create a loss function for tracking loss during testing
    class_weights = torch.Tensor(TrainOptions.CLASS_WEIGHTS)
    # Loss not doing class weights for the first pass
    criterion_base = nn.CrossEntropyLoss()
    criterion_rle = nn.CrossEntropyLoss(weight=class_weights)

    # if gpu is true then transfer the loss over cuda
    if gpu_mode is True:
        criterion_base = criterion_base.cuda()
        criterion_rle = criterion_rle.cuda()

    # initialize base and rle confusion matrix
    sys.stderr.write(TextColor.PURPLE + 'Test starting\n' + TextColor.END)
    base_confusion_matrix = meter.ConfusionMeter(num_base_classes)
    rle_confusion_matrix = meter.ConfusionMeter(num_rle_classes)

    # initialize the accuracy matrices
    total_loss = 0
    total_loss_rle = 0
    total_images = 0
    accuracy = 0

    with torch.no_grad():
        with tqdm(total=len(test_loader), desc='Accuracy: ', leave=True, ncols=100) as pbar:
            # iterate over the dataset in minibatch
            for ii, (images, label_base, label_rle) in enumerate(test_loader):
                # convert the tensors to a proper datatype
                images = images.type(torch.FloatTensor)
                label_base = label_base.type(torch.LongTensor)
                label_rle = label_rle.type(torch.LongTensor)

                if gpu_mode:
                    # encoder_hidden = encoder_hidden.cuda()
                    images = images.cuda()
                    label_base = label_base.cuda()
                    label_rle = label_rle.cuda()

                # initialize the hidden input for the first chunk
                hidden = torch.zeros(images.size(0), 2 * TrainOptions.GRU_LAYERS, TrainOptions.HIDDEN_SIZE)

                if gpu_mode:
                    hidden = hidden.cuda()

                # slide over the image in a sliding window manner
                for i in range(0, ImageSizeOptions.SEQ_LENGTH, TrainOptions.WINDOW_JUMP):
                    # if current position + window size goes beyond the size of the window,
                    # that means we've reached the end
                    if i + TrainOptions.TRAIN_WINDOW > ImageSizeOptions.SEQ_LENGTH:
                        break

                    # chunk the images and labels to this window
                    image_chunk = images[:, i:i+TrainOptions.TRAIN_WINDOW]
                    label_base_chunk = label_base[:, i:i+TrainOptions.TRAIN_WINDOW]
                    label_rle_chunk = label_rle[:, i:i+TrainOptions.TRAIN_WINDOW]

                    # perform an inference on the images
                    output_base, output_rle, hidden = transducer_model(image_chunk, hidden)

                    # calculate loss between the prediction and the true labels
                    loss_base = criterion_base(output_base.contiguous().view(-1, num_base_classes),
                                               label_base_chunk.contiguous().view(-1))
                    loss_rle = criterion_rle(output_rle.contiguous().view(-1, num_rle_classes),
                                             label_rle_chunk.contiguous().view(-1))

                    loss = loss_base + loss_rle

                    # populate the confusion matrix
                    base_confusion_matrix.add(output_base.data.contiguous().view(-1, num_base_classes),
                                              label_base_chunk.data.contiguous().view(-1))
                    rle_confusion_matrix.add(output_rle.data.contiguous().view(-1, num_rle_classes),
                                             label_rle_chunk.data.contiguous().view(-1))

                    total_loss += loss.item()
                    total_images += images.size(0)
                    total_loss_rle += loss_rle.item()

                pbar.update(1)
                # we calculate the accuracy using the confusion matrix
                base_cm_value = base_confusion_matrix.value()
                rle_cm_value = rle_confusion_matrix.value()
                # the sum of all cells in the confusion matrix is the denominator
                base_denom = base_cm_value.sum()
                rle_denom = rle_cm_value.sum()

                # calculate the correct predictions in each case
                base_corrects = 0
                for label in range(0, ImageSizeOptions.TOTAL_BASE_LABELS):
                    base_corrects = base_corrects + base_cm_value[label][label]

                rle_corrects = 0
                for label in range(0, ImageSizeOptions.TOTAL_RLE_LABELS):
                    rle_corrects = rle_corrects + rle_cm_value[label][label]

                # calculate the accuracy
                base_accuracy = 100.0 * (base_corrects / max(1.0, base_denom))
                rle_accuracy = 100.0 * (rle_corrects / max(1.0, rle_denom))

                # set the tqdm bar's accuracy and loss value
                pbar.set_description("Base acc: " + str(round(base_accuracy, 4)) +
                                     ", RLE acc: " + str(round(rle_accuracy, 4)) +
                                     ", RLE loss: " + str(round(total_loss_rle, 4)))

    avg_loss = total_loss / total_images if total_images else 0
    np.set_printoptions(threshold=np.inf)
    # print some statistics
    sys.stderr.write(TextColor.YELLOW+'\nTest Loss: ' + str(avg_loss) + "\n"+TextColor.END)
    sys.stderr.write(TextColor.BLUE + "Base Confusion Matrix: \n" + str(base_confusion_matrix.value()) + "\n" + TextColor.END)
    sys.stderr.write(TextColor.RED + "RLE Confusion Matrix: \n" + TextColor.END)
    for row in rle_confusion_matrix.value():
        row = row.tolist()
        for elem in row:
            sys.stderr.write("{:9d} ".format(elem))
        sys.stderr.write("\n")

    return {'loss': avg_loss, 'accuracy': accuracy, 'base_confusion_matrix': base_confusion_matrix.conf,
            'rle_confusion_matrix': rle_confusion_matrix.conf}



import sys
import torch
import torch.nn as nn
import os
from tqdm import tqdm

from torch.utils.data import DataLoader
from modules.python.models.dataloader import SequenceDataset
from modules.python.TextColor import TextColor
from modules.python.models.test import test
from modules.python.Options import ImageSizeOptions, TrainOptions
from modules.python.models.ModelHander import ModelHandler
"""
This script implements the train method of HELEN.
The train method trains a model on a given set of images and saves model files after each epoch.
"""


def train(train_file,
          test_file,
          batch_size,
          epoch_limit,
          gpu_mode,
          num_workers,
          retrain_model,
          retrain_model_path,
          gru_layers,
          hidden_size,
          lr,
          decay,
          model_dir,
          stats_dir,
          not_hyperband):
    """
    This method implements the training scheme of HELEN. It takes a set of training images and a set of testing
    images and trains the model on the training images and after each epoch it evaluates the trained model on the
    test image set. It saves all model state after each epoch regardless of it's performance. It also saves some
    statistics of the saved models and the training run.
    :param train_file: Path to train image set.
    :param test_file: Path to test image set.
    :param batch_size: Batch size for minibatch operation
    :param epoch_limit: Number of iterations the training will go for
    :param gpu_mode: If True, training and testing will be done on GPU
    :param num_workers: Number of workers for dataloader
    :param retrain_model: If True then it will load a previously-trained model for retraining
    :param retrain_model_path: Path to a previously trained model
    :param gru_layers: Number of GRU layers in the model
    :param hidden_size: Hidden size of the model
    :param lr: Learning Rate for the optimizer
    :param decay: Weight Decay for the optimizer
    :param model_dir: Directory where models will be saved
    :param stats_dir: Directory where statistics of this run will be saved
    :param not_hyperband: This is used by hyperband. If True then hyperband is not running.
    :return:
    """

    # if hyperband is not running then create stat logger for train, test and confusion
    if not_hyperband is True:
        train_loss_logger = open(stats_dir + "train_loss.csv", 'w')
        test_loss_logger = open(stats_dir + "test_loss.csv", 'w')
        confusion_matrix_logger = open(stats_dir + "confusion_matrix.txt", 'w')
    else:
        train_loss_logger = None
        test_loss_logger = None
        confusion_matrix_logger = None

    sys.stderr.write(TextColor.PURPLE + 'Loading data\n' + TextColor.END)
    # initialize training dataset loader
    train_data_set = SequenceDataset(train_file)
    train_loader = DataLoader(train_data_set,
                              batch_size=batch_size,
                              shuffle=True,
                              num_workers=num_workers,
                              pin_memory=gpu_mode)
    num_base_classes = ImageSizeOptions.TOTAL_BASE_LABELS
    num_rle_classes = ImageSizeOptions.TOTAL_RLE_LABELS

    # if retrain model is true then load the model from the model path
    if retrain_model is True:
        if os.path.isfile(retrain_model_path) is False:
            sys.stderr.write(TextColor.RED + "ERROR: INVALID PATH TO RETRAIN PATH MODEL --retrain_model_path\n")
            exit(1)
        sys.stderr.write(TextColor.GREEN + "INFO: RETRAIN MODEL LOADING\n" + TextColor.END)
        transducer_model, hidden_size, gru_layers, prev_ite = \
            ModelHandler.load_simple_model(retrain_model_path,
                                           input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                           image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                           seq_len=ImageSizeOptions.SEQ_LENGTH,
                                           num_base_classes=num_base_classes,
                                           num_rle_classes=num_rle_classes)

        if not_hyperband is True:
            epoch_limit = prev_ite + epoch_limit

        sys.stderr.write(TextColor.GREEN + "INFO: RETRAIN MODEL LOADED\n" + TextColor.END)
    else:
        # if training from scractch, then create a new model
        transducer_model = ModelHandler.get_new_gru_model(input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                                          image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                                          gru_layers=gru_layers,
                                                          hidden_size=hidden_size,
                                                          num_base_classes=num_base_classes,
                                                          num_rle_classes=num_rle_classes)
        prev_ite = 0

    # count the number of trainable parameters for reporting
    param_count = sum(p.numel() for p in transducer_model.parameters() if p.requires_grad)
    sys.stderr.write(TextColor.RED + "INFO: TOTAL TRAINABLE PARAMETERS:\t" + str(param_count) + "\n" + TextColor.END)

    # create a model optimizer
    model_optimizer = torch.optim.Adam(transducer_model.parameters(), lr=lr, weight_decay=decay)
    # this learning rate scheduler reduces learning rate when model reaches plateau
    lr_scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(model_optimizer, 'min')

    # if retrain model is true then load the optimizer
    if retrain_model is True:
        sys.stderr.write(TextColor.GREEN + "INFO: OPTIMIZER LOADING\n" + TextColor.END)
        model_optimizer = ModelHandler.load_simple_optimizer(model_optimizer, retrain_model_path, gpu_mode)
        sys.stderr.write(TextColor.GREEN + "INFO: OPTIMIZER LOADED\n" + TextColor.END)

    class_weights = torch.Tensor(TrainOptions.CLASS_WEIGHTS)
    # we perform a multi-task classification, so we need two loss functions, each performing a single task
    # criterion base is the loss function for base prediction
    criterion_base = nn.CrossEntropyLoss()
    # criterion rle is the loss function for RLE prediction
    criterion_rle = nn.CrossEntropyLoss(weight=class_weights)

    # if gpu mode is true then transfer the model and loss functions to cuda
    if gpu_mode is True:
        transducer_model = torch.nn.DataParallel(transducer_model).cuda()
        criterion_base = criterion_base.cuda()
        criterion_rle = criterion_rle.cuda()

    start_epoch = prev_ite

    # Train the Model
    sys.stderr.write(TextColor.PURPLE + 'Training starting\n' + TextColor.END)
    # create stats dicts
    stats = dict()
    stats['loss_epoch'] = []
    stats['accuracy_epoch'] = []
    sys.stderr.write(TextColor.BLUE + 'Start: ' + str(start_epoch + 1) + ' End: ' + str(epoch_limit) + "\n")

    # for each epoch we iterate over the training dataset once
    for epoch in range(start_epoch, epoch_limit, 1):
        total_loss_base = 0
        total_loss_rle = 0
        total_loss = 0
        total_images = 0
        sys.stderr.write(TextColor.BLUE + 'Train epoch: ' + str(epoch + 1) + "\n")
        batch_no = 1

        # tqdm is the progress bar we use for logging
        with tqdm(total=len(train_loader), desc='Loss', leave=True, ncols=100) as progress_bar:
            # make sure the model is in train mode. BN is different in train and eval.
            transducer_model.train()
            for images, label_base, label_rle in train_loader:
                # convert the tensors to the proper datatypes.
                images = images.type(torch.FloatTensor)
                label_base = label_base.type(torch.LongTensor)
                label_rle = label_rle.type(torch.LongTensor)

                # initialize the hidden input for the first chunk
                hidden = torch.zeros(images.size(0), 2 * TrainOptions.GRU_LAYERS, TrainOptions.HIDDEN_SIZE)

                # if gpu_mode is true then transfer all tensors to cuda
                if gpu_mode:
                    images = images.cuda()
                    label_base = label_base.cuda()
                    label_rle = label_rle.cuda()
                    hidden = hidden.cuda()

                # perform a sliding window on the entire image sequence length
                for i in range(0, ImageSizeOptions.SEQ_LENGTH, TrainOptions.WINDOW_JUMP):
                    # we optimize over each chunk
                    model_optimizer.zero_grad()

                    # if current position + window size goes beyond the size of the window,
                    # that means we've reached the end
                    if i + TrainOptions.TRAIN_WINDOW > ImageSizeOptions.SEQ_LENGTH:
                        break

                    # get the chunks for this window
                    image_chunk = images[:, i:i+TrainOptions.TRAIN_WINDOW]
                    label_base_chunk = label_base[:, i:i+TrainOptions.TRAIN_WINDOW]
                    label_rle_chunk = label_rle[:, i:i+TrainOptions.TRAIN_WINDOW]

                    # get the inference from the model
                    output_base, output_rle, hidden = transducer_model(image_chunk, hidden)

                    # calculate loss for base prediction
                    loss_base = criterion_base(output_base.contiguous().view(-1, num_base_classes),
                                               label_base_chunk.contiguous().view(-1))
                    # calculate loss for RLE prediction
                    loss_rle = criterion_rle(output_rle.contiguous().view(-1, num_rle_classes),
                                             label_rle_chunk.contiguous().view(-1))

                    # sum the losses to have a singlee optimization over multiple tasks
                    loss = loss_base + loss_rle

                    # backpropagation and weight update
                    loss.backward()
                    model_optimizer.step()

                    # update the loss values
                    total_loss += loss.item()
                    total_loss_base += loss_base.item()
                    total_loss_rle += loss_rle.item()
                    total_images += image_chunk.size(0)

                    # detach the hidden from the graph as the next chunk will be a new optimization
                    hidden = hidden.detach()

                # update the progress bar
                avg_loss = (total_loss / total_images) if total_images else 0
                progress_bar.set_description("Base: " + str(round(total_loss_base, 4)) +
                                             ", RLE: " + str(round(total_loss_rle, 4)) +
                                             ", TOTAL: " + str(round(total_loss, 4)))

                if not_hyperband is True:
                    train_loss_logger.write(str(epoch + 1) + "," + str(batch_no) + "," + str(avg_loss) + "\n")
                progress_bar.refresh()
                progress_bar.update(1)
                batch_no += 1

            progress_bar.close()

        # after each epoch, evaluate the current state of the model
        stats_dictionary = test(test_file, batch_size, gpu_mode, transducer_model, num_workers,
                                gru_layers, hidden_size, num_base_classes=ImageSizeOptions.TOTAL_BASE_LABELS,
                                num_rle_classes=ImageSizeOptions.TOTAL_RLE_LABELS)
        stats['loss'] = stats_dictionary['loss']
        stats['accuracy'] = stats_dictionary['accuracy']
        stats['loss_epoch'].append((epoch, stats_dictionary['loss']))
        stats['accuracy_epoch'].append((epoch, stats_dictionary['accuracy']))

        lr_scheduler.step(stats['loss'])

        # save the model after each epoch and update the loggers after each epoch
        if not_hyperband is True:
            ModelHandler.save_model(transducer_model, model_optimizer,
                                    hidden_size, gru_layers,
                                    epoch, model_dir + "HELEN_epoch_" + str(epoch + 1) + '_checkpoint.pkl')
            sys.stderr.write(TextColor.RED + "\nMODEL SAVED SUCCESSFULLY.\n" + TextColor.END)

            test_loss_logger.write(str(epoch + 1) + "," + str(stats['loss']) + "," + str(stats['accuracy']) + "\n")
            confusion_matrix_logger.write(str(epoch + 1) + "\n" + str(stats_dictionary['base_confusion_matrix']) + "\n")
            train_loss_logger.flush()
            test_loss_logger.flush()
            confusion_matrix_logger.flush()
        # else:
        #     # this setup is for hyperband
        #     if epoch + 1 >= 2 and stats['accuracy'] < 90:
        #         sys.stderr.write(TextColor.PURPLE + 'EARLY STOPPING AS THE MODEL NOT DOING WELL\n' + TextColor.END)
        #         return transducer_model, model_optimizer, stats

    # notify that the model has finished training.
    sys.stderr.write(TextColor.PURPLE + 'Finished training\n' + TextColor.END)




from __future__ import print_function
import numpy as np
from math import log, ceil
from time import time, ctime
import sys
from modules.python.TextColor import TextColor
import logging
from datetime import datetime
"""
The Hyperband class used for hyper-parameter optimization.
Optimized from:
https://github.com/zygmuntz/hyperband
"""


class Hyperband:
    """
    Hyper-parameter optimization algorithm implemented
    This class is optimized from the popular hyperband implementation:
    https://github.com/zygmuntz/hyperband
    """
    def __init__(self, get_params_function, try_params_function, max_iteration, downsample_rate, log_directory,
                 model_directory):
        """
        Initialize hyperband object.
        :param get_params_function: A function to get parameters from sample space.
        :param try_params_function: Try a parameter by training and testing it
        :param max_iteration: Maximum iterations per configuration
        :param downsample_rate: Defines configuration downsampling rate
        :param log_directory: Directory where log is saved
        :param model_directory: Directory where model is saved
        """
        self.get_params = get_params_function
        self.try_params = try_params_function

        self.max_iter = max_iteration  # maximum iterations per configuration
        self.eta = downsample_rate  # defines configuration downsampling rate (default = 3)

        self.logeta = lambda x: log(x) / log(self.eta)
        self.s_max = int(self.logeta(self.max_iter))
        self.B = (self.s_max + 1) * self.max_iter

        self.results = []  # list of dicts
        self.counter = 0
        self.best_loss = np.inf
        self.best_acc = 0
        self.best_counter = -1
        self.best_config = ''
        self.log_file = log_directory+'Hyperband_'+datetime.now().strftime("%Y%m%d_%H%M%S")+'.log'
        self.model_dir = model_directory + 'Hyperband_' + datetime.now().strftime("%Y%m%d_%H%M%S") + "_"

        logging.basicConfig(filename=self.log_file, level=logging.INFO)

    # can be called multiple times
    def run(self, skip_last=0):
        """
        The hyper-parameter optimization algorithm.
        :param skip_last: Skip the last iteration
        :return:
        """
        model_id = 0
        for s in reversed(range(self.s_max + 1)):

            # initial number of configurations
            n = int(ceil(self.B / self.max_iter / (s + 1) * self.eta ** s))

            # initial number of iterations per config
            r = self.max_iter * self.eta ** (-s)

            # n random configurations
            model_configs = [(self.get_params(), False, self.model_dir + 'con_' + str(model_id + i) + '.pkl', 0) for i in range(n)]
            model_id += n
            for i in range((s + 1) - int(skip_last)):  # changed from s + 1

                # Run each of the n configs for <iterations>
                # and keep best (n_configs / eta) configurations

                n_configs = n * self.eta ** (-i)
                n_iterations = int(ceil(r * self.eta ** (i)))

                sys.stderr.write(TextColor.BLUE + "\n*** {} configurations x {:.5f} iterations each"
                                 .format(n_configs, n_iterations) + "\n" + TextColor.END)

                logging.info("\n*** {} configurations x {:.1f} iterations each".format(n_configs, n_iterations))

                val_losses = []
                early_stops = []

                for config_index, config in enumerate(model_configs):
                    self.counter += 1
                    sys.stderr.write(TextColor.BLUE + "{} | {} | lowest loss: {} | accuracy: {} | (run {}) | model {}"
                                     .format(self.counter, ctime(), self.best_loss, self.best_acc, self.best_counter,
                                             self.best_config)
                                     + TextColor.END)
                    logging.info("{} | {} | lowest loss so far: {} | (run {})"
                                 .format(self.counter, ctime(), self.best_loss, self.best_counter))

                    start_time = time()

                    logging.info("Iterations:\t" + str(n_iterations))
                    logging.info("Params:\t" + str(config[0]))
                    params, retrain_model, model_path, prev_ite = config
                    transducer_model, model_optimizer, result = self.try_params(n_iterations, config, model_path)

                    assert (type(result) == dict)
                    assert ('loss' in result)

                    seconds = int(round(time() - start_time))
                    sys.stderr.write(TextColor.BLUE + "\n{} seconds.\n".format(seconds) + TextColor.END)

                    loss = result['loss']
                    val_losses.append(loss)

                    early_stop = result.get('early_stop', False)
                    early_stops.append(early_stop)

                    # keeping track of the best result so far (for display only)
                    # could do it be checking results each time, but hey
                    if loss < self.best_loss:
                        self.best_loss = loss
                        self.best_counter = self.counter
                        self.best_acc = result['accuracy']
                        self.best_config = model_path

                    model_configs[config_index] = (params, True, model_path, n_iterations)

                    result['counter'] = self.counter
                    result['seconds'] = seconds
                    result['params'] = params
                    result['model_path'] = model_path
                    result['iterations'] = n_iterations

                    self.results.append(result)

                # select a number of best configurations for the next loop
                # filter out early stops, if any
                indices = np.argsort(val_losses)
                model_configs = [model_configs[i] for i in indices if not early_stops[i]]
                model_configs = model_configs[0:int(n_configs / self.eta)]

        return self.results



class TextColor:
    """
    Defines color codes for text used to give different mode of errors.
    """
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


import h5py
import sys
import os
from os.path import isfile, join
from os import listdir
from modules.python.Stitch import Stitch
from modules.python.TextColor import TextColor
from modules.python.FileManager import FileManager
"""
The stitch module generates a consensus sequence from all the predictions we generated from call_consensus.py.

The method is as follows:
    1) MarginPolish
     - MarginPolish takes a "Region" which is usually 1000 bases and generates pileup summaries for them.
     - The region often is more than 1000 bases but for mini-batch to work, we make sure all the images are 1000 bases
       so MarginPolish chunks these Regions into images providing each Regional image with a chunk id.
    2) HELEN call consensus
     - Call Consensus generates a prediction for each of the images, but while saving it makes sure all the images
       that belong to the same region, gets saved under the same chunk_prefix, making sure all the sequences can
     - be aggregated easily.
    3) HELEN stitch
     -  The stitch method that loads one contig at a time and grabs all the "regions" with predictions. The regional
        chunks are easily stitched without having to do an overlap as the positions inside a region is consistent.
     - For regions that are adjacent, HELEN uses a local smith-waterman alignment to find an anchor point to
       stitch sequences from two adjacent sequences together.
"""


def get_file_paths_from_directory(directory_path):
    """
    Returns all paths of files given a directory path
    :param directory_path: Path to the directory
    :return: A list of paths of files
    """
    file_paths = [os.path.abspath(join(directory_path, file)) for file in listdir(directory_path)
                  if isfile(join(directory_path, file)) and file[-3:] == 'hdf']
    return file_paths


def perform_stitch(input_directory, output_path, output_prefix, threads):
    """
    This method gathers all contigs and calls the stitch module for each contig.
    :param input_directory: Path to the directory containing input files.
    :param output_path: Path to the output_consensus_sequence
    :param output_prefix: Output file's prefix
    :param threads: Number of threads to use
    :return:
    """
    # get all the files
    all_prediction_files = get_file_paths_from_directory(input_directory)

    # we gather all the contigs
    all_contigs = set()

    # get contigs from all of the files
    for prediction_file in sorted(all_prediction_files):
        with h5py.File(prediction_file, 'r') as hdf5_file:
            if 'predictions' in hdf5_file:
                contigs = list(hdf5_file['predictions'].keys())
                all_contigs.update(contigs)
            else:
                raise ValueError(TextColor.RED + "ERROR: INVALID HDF5 FILE, FILE DOES NOT CONTAIN predictions KEY.\n"
                                 + TextColor.END)
    # convert set to a list
    all_contigs = list(all_contigs)

    # get output directory
    output_dir = FileManager.handle_output_directory(output_path)

    # open an output fasta file
    # we should really use a fasta handler for this, I don't like this.
    output_filename = os.path.join(output_dir, output_prefix + '.fa')
    consensus_fasta_file = open(output_filename, 'w')
    sys.stderr.write(TextColor.GREEN + "INFO: OUTPUT FILE: " + output_filename + "\n" + TextColor.END)

    # for each contig
    for i, contig in enumerate(sorted(all_contigs)):
        log_prefix = "{:04d}".format(i) + "/" + "{:04d}".format(len(contigs)) + ":"
        sys.stderr.write(TextColor.GREEN + "INFO: " + str(log_prefix) + " PROCESSING CONTIG: " + contig + "\n"
                         + TextColor.END)

        # get all the chunk keys
        chunk_name_tuple = list()
        for prediction_file in all_prediction_files:
            with h5py.File(prediction_file, 'r') as hdf5_file:
                # check if the contig is contained in this file
                if contig not in list(hdf5_file['predictions'].keys()):
                    continue

                # if contained then get the chunks
                chunk_keys = sorted(hdf5_file['predictions'][contig].keys())
                for chunk_key in chunk_keys:
                    chunk_contig_start = hdf5_file['predictions'][contig][chunk_key]['contig_start'][()]
                    chunk_contig_end = hdf5_file['predictions'][contig][chunk_key]['contig_end'][()]
                    chunk_name_tuple.append((prediction_file, chunk_key, chunk_contig_start, chunk_contig_end))

        # call stitch to generate a sequence for this contig
        stich_object = Stitch()
        consensus_sequence = stich_object.create_consensus_sequence(contig, chunk_name_tuple, threads)
        sys.stderr.write(TextColor.BLUE + "INFO: " + str(log_prefix) + " FINISHED PROCESSING " + contig
                         + ", POLISHED SEQUENCE LENGTH: " + str(len(consensus_sequence)) + ".\n" + TextColor.END)

        # if theres a sequence then write it to the file
        if consensus_sequence is not None:
            consensus_fasta_file.write('>' + contig + "\n")
            consensus_fasta_file.write(consensus_sequence+"\n")


import argparse
import sys
import torch
from version import __version__
from modules.python.TextColor import TextColor
from modules.python.PolishInterface import polish_genome
from modules.python.CallConsensusInterface import call_consensus
from modules.python.StitchInterface import perform_stitch
from modules.python.DownloadModel import download_models


def add_polish_arguments(parser):
    """
    Add arguments to a parser for sub-command "polish"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "-i",
        "--image_dir",
        type=str,
        required=True,
        help="[REQUIRED] Path to a directory where all MarginPolish generated images are."
    )
    parser.add_argument(
        "-m",
        "--model_path",
        type=str,
        required=True,
        help="[REQUIRED] Path to a trained model (pkl file). Please see our github page to see options."
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        required=False,
        default=512,
        help="Batch size for testing, default is 512. Please set to 512 or 1024 for a balanced execution time."
    )
    parser.add_argument(
        "-w",
        "--num_workers",
        type=int,
        required=False,
        default=8,
        help="Number of workers to assign to the dataloader."
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        required=False,
        default=1,
        help="Number of PyTorch threads to use, default is 1. This may be helpful during CPU-only inference."
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=False,
        default='./output/',
        help="Path to the output directory."
    )
    parser.add_argument(
        "-p",
        "--output_prefix",
        type=str,
        required=False,
        default="HELEN_prediction",
        help="Prefix for the output file. Default is: HELEN_prediction"
    )
    parser.add_argument(
        "-g",
        "--gpu_mode",
        default=False,
        action='store_true',
        help="If set then PyTorch will use GPUs for inference. CUDA required."
    )
    parser.add_argument(
        "-d_ids",
        "--device_ids",
        type=str,
        required=False,
        default=None,
        help="List of gpu device ids to use for inference. Only used in distributed setting.\n"
             "Example usage: --device_ids 0,1,2 (this will create three callers in id 'cuda:0, cuda:1 and cuda:2'\n"
             "If none then it will use all available devices."
    )
    parser.add_argument(
        "-c",
        "--callers",
        type=int,
        required=False,
        default=8,
        help="Total number of callers to spawn if doing CPU inference in distributed mode."
    )
    return parser


def add_call_consensus_arguments(parser):
    """
    Add arguments to a parser for sub-command "call_consensus"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "-i",
        "--image_dir",
        type=str,
        required=True,
        help="[REQUIRED] Path to a directory where all MarginPolish generated images are."
    )
    parser.add_argument(
        "-m",
        "--model_path",
        type=str,
        required=True,
        help="[REQUIRED] Path to a trained model (pkl file). Please see our github page to see options."
    )
    parser.add_argument(
        "-b",
        "--batch_size",
        type=int,
        required=False,
        default=512,
        help="Batch size for testing, default is 512. Please set to 512 or 1024 for a balanced execution time."
    )
    parser.add_argument(
        "-w",
        "--num_workers",
        type=int,
        required=False,
        default=8,
        help="Number of workers to assign to the dataloader. Shouldg be 0 if using Docker."
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=False,
        default='./output/',
        help="Path to the output directory."
    )
    parser.add_argument(
        "-p",
        "--output_prefix",
        type=str,
        required=False,
        default="HELEN_prediction",
        help="Prefix for the output file. Default is: HELEN_prediction"
    )
    parser.add_argument(
        "-g",
        "--gpu_mode",
        default=False,
        action='store_true',
        help="If set then PyTorch will use GPUs for inference. CUDA required."
    )
    parser.add_argument(
        "-d_ids",
        "--device_ids",
        type=str,
        required=False,
        default=None,
        help="List of gpu device ids to use for inference. Only used in distributed setting.\n"
             "Example usage: --device_ids 0,1,2 (this will create three callers in id 'cuda:0, cuda:1 and cuda:2'\n"
             "If none then it will use all available devices."
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        required=False,
        default=16,
        help="Total available threads to use."
    )
    parser.add_argument(
        "-c",
        "--callers",
        type=int,
        required=False,
        default=8,
        help="Total number of callers to spawn if doing CPU inference in distributed mode."
    )
    return parser


def add_stitch_arguments(parser):
    """
    Add arguments to a parser for sub-command "stitch"
    :param parser: argeparse object
    :return:
    """
    parser.add_argument(
        "-i",
        "--input_dir",
        type=str,
        required=True,
        help="[REQUIRED] Path to a directory containing prediction files call consensus."
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=True,
        help="[REQUIRED] Path to the output directory."
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=int,
        required=True,
        help="[REQUIRED] Number of threads."
    )
    parser.add_argument(
        "-p",
        "--output_prefix",
        type=str,
        required=False,
        default="HELEN_consensus",
        help="Prefix for the output file. Default is: HELEN_consensus"
    )


def add_download_models_arguments(parser):
    """
    Add parameters for model download script
    :param parser: Parser object
    :return:
    """
    parser.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=True,
        help="Path to directory where models will be saved."
    )
    return parser


def main():
    """
    Main interface for HELEN. The submodules supported as of now are these:
    1) Polish
    2) Call_consensus
    3) Stitch
    """
    parser = argparse.ArgumentParser(description="HELEN is a RNN based polisher for polishing ONT-based assemblies. "
                                                 "You can avail three commands with this script:\n"
                                                 "1) polish: Call consensus and stitch the predictions.\n"
                                                 "2) call_consensus: This module takes the summary images and a"
                                                 "trained neural network and generates predictions per base.\n"
                                                 "3) stitch: This module takes the inference files as input and "
                                                 "stitches them to generate a polished assembly.\n"
                                                 "4) download_model: Download available helen models\n"
                                                 "5) torch_stat: See the torch configuration\n"
                                                 "6) version: check HELEN version.\n",

                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "--version",
        default=False,
        action='store_true',
        help="Show version."
    )

    subparsers = parser.add_subparsers(dest='sub_command')
    # subparsers.required = True

    parser_polish = subparsers.add_parser('polish', help="Run the polishing pipeline. This will run "
                                                         "call_consensus -> stitch one after another.\n"
                                                         "The outputs of each step can be run separately using\n"
                                                         "the appropriate sub-command.")
    add_polish_arguments(parser_polish)

    parser_call_consensus = subparsers.add_parser('call_consensus',
                                                  help="call_consensus.py script HELEN performs inference on a given "
                                                       "set of images generated by MarginPolish. "
                                                       "The input of this script is a directory containing all the"
                                                       " images generated by MarginPolish, a model and an "
                                                       "output directory.\n"
                                                       "OUTPUT: This module generates .hdf files which is used by the "
                                                       "stitch.py script to perform the final stitching.")
    add_call_consensus_arguments(parser_call_consensus)

    parser_stitch = subparsers.add_parser('stitch',
                                          help="stitch.py of HELEN generates a polished sequence from a given hdf"
                                               " file generated by call_consensus.py script. \nstitch.py is "
                                               " multi-threaded and faster run-time can be achieved on higer number"
                                               " of CPUs.\n"
                                               "OUTPUT: Is a FASTA file containing a polished consensus sequence.")
    add_stitch_arguments(parser_stitch)

    parser_download_model = subparsers.add_parser('download_models', help="Download available models.")
    add_download_models_arguments(parser_download_model)

    parser_torch_stat = subparsers.add_parser('torch_stat', help="See PyTorch configuration.")
    parser_torch_stat = subparsers.add_parser('version', help="Show program version.")

    FLAGS, unparsed = parser.parse_known_args()

    if FLAGS.sub_command == 'polish':
        sys.stderr.write(TextColor.GREEN + "INFO: POLISH MODULE SELECTED\n" + TextColor.END)
        polish_genome(FLAGS.image_dir,
                      FLAGS.model_path,
                      FLAGS.batch_size,
                      FLAGS.num_workers,
                      FLAGS.threads,
                      FLAGS.output_dir,
                      FLAGS.output_prefix,
                      FLAGS.gpu_mode,
                      FLAGS.device_ids,
                      FLAGS.callers)

    elif FLAGS.sub_command == 'call_consensus':
        sys.stderr.write(TextColor.GREEN + "INFO: CALL CONSENSUS MODULE SELECTED\n" + TextColor.END)
        call_consensus(FLAGS.image_dir,
                       FLAGS.model_path,
                       FLAGS.batch_size,
                       FLAGS.num_workers,
                       FLAGS.threads,
                       FLAGS.output_dir,
                       FLAGS.output_prefix,
                       FLAGS.gpu_mode,
                       FLAGS.device_ids,
                       FLAGS.callers)

    elif FLAGS.sub_command == 'stitch':
        sys.stderr.write(TextColor.GREEN + "INFO: STITCH MODULE SELECTED\n" + TextColor.END)
        perform_stitch(FLAGS.input_dir,
                       FLAGS.output_dir,
                       FLAGS.output_prefix,
                       FLAGS.threads)

    elif FLAGS.sub_command == 'download_models':
        sys.stderr.write(TextColor.GREEN + "INFO: DOWNLOAD MODELS SELECTED\n" + TextColor.END)
        download_models(FLAGS.output_dir)

    elif FLAGS.sub_command == 'torch_stat':
        sys.stderr.write(TextColor.YELLOW + "TORCH VERSION: " + TextColor.END + str(torch.__version__) + "\n\n")
        sys.stderr.write(TextColor.YELLOW + "PARALLEL CONFIG:\n" + TextColor.END)
        print(torch.__config__.parallel_info())
        sys.stderr.write(TextColor.YELLOW + "BUILD CONFIG:\n" + TextColor.END)
        print(*torch.__config__.show().split("\n"), sep="\n")

        sys.stderr.write(TextColor.GREEN + "CUDA AVAILABLE: " + TextColor.END + str(torch.cuda.is_available()) + "\n")
        sys.stderr.write(TextColor.GREEN + "GPU DEVICES: " + TextColor.END + str(torch.cuda.device_count()) + "\n")

    elif FLAGS.version is True:
        print("HELEN VERSION: ", __version__)

    else:
        sys.stderr.write(TextColor.RED + "ERROR: NO SUBCOMMAND SELECTED. "
                                         "PLEASE SELECT ONE OF THE AVAIABLE SUB-COMMANDS.\n"
                         + TextColor.END)
        parser.print_help()


if __name__ == '__main__':
    main()


import numpy as np
from torch.utils.data import Dataset
import torchvision.transforms as transforms
import h5py
import sys
from modules.python.TextColor import TextColor
from modules.python.Options import ImageSizeOptions
from modules.python.FileManager import FileManager


class SequenceDataset(Dataset):
    """
    This class implements the dataset class for the dataloader to use.
    This version is intended to use with predict.py.
    It initializes all the given images and returns each image through __getitem__.
    """

    def __init__(self, image_directory, file_list=None):
        """
        This method initializes the dataset by loading all the image information. It creates a sequential list
        call all_images from which we can grab images iteratively through __getitem__.
        :param image_directory: Path to a directory where all the images are saved.
        """
        # transformer to convert loaded objects to tensors
        self.transform = transforms.Compose([transforms.ToTensor()])

        # a list of file-image pairs, where we have (file_name, image_name) as values so we can fetch images
        # from the list of files.
        file_image_pair = []

        # check if a file list is given, otherwise load all files from the directory
        if file_list is not None:
            hdf_files = file_list
        else:
            # get all the h5 files that we have in the directory
            hdf_files = FileManager.get_file_paths_from_directory(image_directory)

        for hdf5_file_path in hdf_files:
            # for each of the files get all the images
            with h5py.File(hdf5_file_path, 'r') as hdf5_file:
                if 'images' in hdf5_file:
                    image_names = list(hdf5_file['images'].keys())

                    # save the file-image pair to the list
                    for image_name in image_names:
                        file_image_pair.append((hdf5_file_path, image_name))
                else:
                    sys.stderr.write(TextColor.YELLOW + "WARN: NO IMAGES FOUND IN FILE: "
                                     + hdf5_file_path + "\n" + TextColor.END)

        # save the list to all_images so we can access the list inside other methods
        self.all_images = file_image_pair

    def __getitem__(self, index):
        """
        This method returns a single object. Dataloader uses this method to load images and then minibatches the loaded
        images
        :param index: Index indicating which image from all_images to be loaded
        :return: image and their auxiliary information
        """
        hdf5_filepath, image_name = self.all_images[index]

        # load all the information we need to save in the prediction hdf5
        with h5py.File(hdf5_filepath, 'r') as hdf5_file:
            contig = np.array2string(hdf5_file['images'][image_name]['contig'][()][0].astype(np.str)).replace("'", '')
            contig_start = hdf5_file['images'][image_name]['contig_start'][()][0].astype(np.int)
            contig_end = hdf5_file['images'][image_name]['contig_end'][()][0].astype(np.int)
            chunk_id = hdf5_file['images'][image_name]['feature_chunk_idx'][()][0].astype(np.int)
            image = hdf5_file['images'][image_name]['image'][()].astype(np.uint8)
            position = hdf5_file['images'][image_name]['position'][()].astype(np.int)

        # if the size of the image is smaller than the sequence length, then we need to pad to the image to make
        # it to image size.
        if image.shape[0] < ImageSizeOptions.SEQ_LENGTH:
            total_empty_needed = ImageSizeOptions.SEQ_LENGTH - image.shape[0]
            empty_image_columns = np.array([[0] * ImageSizeOptions.IMAGE_HEIGHT] * total_empty_needed)
            image = np.append(image, empty_image_columns, 0)
            image = image.astype(np.uint8)

            empty_positions = np.array([[-1, -1, -1]] * total_empty_needed)
            position = np.append(position, empty_positions, 0)
            position = position.astype(np.int)

        # at this point the image size should be SEQ_LENGTH, if not then raise a ValueError.
        if image.shape[0] < ImageSizeOptions.SEQ_LENGTH or position.shape[0] < ImageSizeOptions.SEQ_LENGTH:
            raise ValueError("IMAGE SIZE ERROR: " + str(hdf5_filepath) + " " + str(image.shape))

        return contig, contig_start, contig_end, chunk_id, image, position, hdf5_filepath

    def __len__(self):
        """
        Returns the length of the dataset
        :return: Int value containing the length of the dataset
        """
        return len(self.all_images)



import sys
import os
import torch
from modules.python.TextColor import TextColor
from modules.python.models.predict_cpu import predict_cpu
from modules.python.models.predict_gpu import predict_gpu
from modules.python.FileManager import FileManager
from os.path import isfile, join
from os import listdir
"""
The Call Consensus method generates base predictions for images generated through MarginPolish. This script reads
hdf5 files generated by MarginPolish and produces another Hdf5 file that holds all predictions. The generated hdf5 file
is given to stitch.py which then stitches the segments using an alignment which gives us a polished sequence.

The algorithm is described here:

  1) INPUTS:
    - directory path to the image files generated by MarginPolish
    - model path directing to a trained model
    - batch size for mini-batch prediction
    - num workers for mini-batch processing threads
    - output directory path to where the output hdf5 will be saved
    - gpu mode indicating if GPU will be used
  2) METHOD:
    - Call predict function that loads the neural network and generates base predictions and saves it into a hdf5 file
        - Loads the model
        - Iterates over the input images in minibatch
        - For each image uses a sliding window method to slide of the image sequence
        - Aggregate the predictions to get sequence prediction for the entire image sequence
        - Save all the predictions to a file
  3) OUTPUT:
    - A hdf5 file containing all the base predictions   
"""


def get_file_paths_from_directory(directory_path):
    """
    Returns all paths of files given a directory path
    :param directory_path: Path to the directory
    :return: A list of paths of files
    """
    file_paths = [os.path.abspath(join(directory_path, file)) for file in listdir(directory_path)
                  if isfile(join(directory_path, file)) and file[-2:] == 'h5']
    return file_paths


def call_consensus(image_dir, model_path, batch_size, num_workers, threads, output_dir, output_prefix, gpu_mode,
                   device_ids, callers):
    """
    This method provides an interface too call the predict method that generates the prediction hdf5 file
    :param image_dir: Path to directory where all MarginPolish images are saved
    :param model_path: Path to a trained model
    :param batch_size: Batch size for minibatch processing
    :param num_workers: Number of workers for minibatch processing
    :param threads: Number of threads for pytorch
    :param output_dir: Path to the output directory
    :param output_prefix: Prefix of the output HDF5 file
    :param gpu_mode: If true, predict method will use GPU.
    :param device_ids: List of CUDA devices to use.
    :param callers: Total number of callers.
    :return:
    """
    # check the model file
    if not os.path.isfile(model_path):
        sys.stderr.write(TextColor.RED + "ERROR: CAN NOT LOCATE MODEL FILE.\n" + TextColor.END)
        exit(1)

    # check the input directory
    if not os.path.isdir(image_dir):
        sys.stderr.write(TextColor.RED + "ERROR: CAN NOT LOCATE IMAGE DIRECTORY.\n" + TextColor.END)
        exit(1)

    # check batch_size
    if batch_size <= 0:
        sys.stderr.write(TextColor.RED + "ERROR: batch_size NEEDS TO BE >0.\n" + TextColor.END)
        exit(1)

    # check num_workers
    if num_workers < 0:
        sys.stderr.write(TextColor.RED + "ERROR: num_workers NEEDS TO BE >=0.\n" + TextColor.END)
        exit(1)

    # check number of threads
    if threads <= 0:
        sys.stderr.write(TextColor.RED + "ERROR: THREAD NEEDS TO BE >=0.\n" + TextColor.END)
        exit(1)

    output_dir = FileManager.handle_output_directory(output_dir)

    # create a filename for the output file
    output_filename = os.path.join(output_dir, output_prefix)

    # inform the output directory
    sys.stderr.write(TextColor.GREEN + "INFO: " + TextColor.END + "OUTPUT FILE: " + output_filename + "\n")

    if gpu_mode:
        # Make sure that GPU is
        if not torch.cuda.is_available():
            sys.stderr.write(TextColor.RED + "ERROR: TORCH IS NOT BUILT WITH CUDA.\n" + TextColor.END)
            sys.stderr.write(TextColor.RED + "SEE TORCH CAPABILITY:\n$ python3\n"
                                             ">>> import torch \n"
                                             ">>> torch.cuda.is_available()\n If true then cuda is avilable"
                             + TextColor.END)
            exit(1)

        # Now see which devices to use
        if device_ids is None:
            total_gpu_devices = torch.cuda.device_count()
            sys.stderr.write(TextColor.GREEN + "INFO: TOTAL GPU AVAILABLE: " + str(total_gpu_devices) + "\n" + TextColor.END)
            device_ids = [i for i in range(0, total_gpu_devices)]
            callers = total_gpu_devices
        else:
            device_ids = [int(i) for i in device_ids.split(',')]
            for device_id in device_ids:
                major_capable, minor_capable = torch.cuda.get_device_capability(device=device_id)
                if major_capable < 0:
                    sys.stderr.write(TextColor.RED + "ERROR: GPU DEVICE: " + str(device_id) + " IS NOT CUDA CAPABLE.\n" + TextColor.END)
                    sys.stderr.write(TextColor.GREEN + "Try running: $ python3\n"
                                                       ">>> import torch \n"
                                                       ">>> torch.cuda.get_device_capability(device="
                                     + str(device_id) + ")\n" + TextColor.END)
                else:
                    sys.stderr.write(TextColor.GREEN + "INFO: CAPABILITY OF GPU#" + str(device_id)
                                     + ":\t" + str(major_capable) + "-" + str(minor_capable) + "\n" + TextColor.END)
            callers = len(device_ids)

        sys.stderr.write(TextColor.GREEN + "INFO: AVAILABLE GPU DEVICES: " + str(device_ids) + "\n" + TextColor.END)
        threads_per_caller = 0
    else:
        # calculate how many threads each caller can use
        threads_per_caller = int(threads / callers)
        device_ids = []

    # chunk the inputs
    input_files = get_file_paths_from_directory(image_dir)

    # generate file chunks to process in parallel
    file_chunks = [[] for i in range(callers)]
    for i in range(0, len(input_files)):
        file_chunks[i % callers].append(input_files[i])

    # get the file chunks
    file_chunks = [file_chunks[i] for i in range(len(file_chunks)) if len(file_chunks[i]) > 0]

    callers = len(file_chunks)

    if gpu_mode:
        # Distributed GPU setup
        predict_gpu(file_chunks, output_filename, model_path, batch_size, callers, device_ids, num_workers)
    else:
        # distributed CPU setup, call the prediction function
        predict_cpu(file_chunks, output_filename, model_path, batch_size,
                    callers, threads_per_caller, num_workers)

    # notify the user that process has completed successfully
    sys.stderr.write(TextColor.GREEN + "INFO: " + TextColor.END + "PREDICTION GENERATED SUCCESSFULLY.\n")



import argparse
"""
HELPER SCRIPT TO COUNT QUAST MISASSEMBLIES THAT HAPPEN OUTSIDE KNOWN SVs.
"""


def read_quast_file(file):
    """
    Extracts misassembly regions from QUAST file, usually located in /QUAST_output/contig_reports/all_alignments_$$.tsv
    :param file:
    :return:
    """
    total_count = 0
    relocation_count = 0
    translocation_count = 0
    inversion_count = 0
    misassemblies = []
    with open(file) as f:
        prev_line = ''
        for line in f:
            line = line.rstrip().replace(',', '')
            splits = line.split('\t')
            if splits[0].split(' ')[0] == 'relocation':
                s_ref, e_ref, s_con, e_con, ref, con, idn, ambi, bg = prev_line.split('\t')
                # print(ref, s_ref, e_ref, splits[0])
                misassemblies.append([ref, s_ref, e_ref, splits[0].split(' ')[0]])
                relocation_count += 1
                total_count += 1
            elif splits[0] == 'translocation':
                s_ref, e_ref, s_con, e_con, ref, con, idn, ambi, bg = prev_line.split('\t')
                # print(ref, s_ref, e_ref, splits[0])
                misassemblies.append([ref, s_ref, e_ref, splits[0]])
                translocation_count += 1
                total_count += 1
            elif splits[0] == 'inversion':
                s_ref, e_ref, s_con, e_con, ref, con, idn, ambi, bg = prev_line.split('\t')
                # print(ref, s_ref, e_ref, splits[0])
                misassemblies.append([ref, s_ref, e_ref, splits[0]])
                inversion_count += 1
                total_count += 1
            prev_line = line

    print("#####--TOTAL MISSASSEMBLIES: REPORTED BY QUAST--#####")
    print("Total Misassemblies:\t", total_count)
    print("Total relocations:\t", relocation_count)
    print("Total translocations:\t", translocation_count)
    print("Total inversions:\t", inversion_count)
    print("#####################################################\n")
    # returns a list of misassemblies as a list of list where each element is [chr_name, st, end, type_of_missassmbly]
    return misassemblies


def read_bed_file(file):
    known_svs = []
    with open(file) as f:
        for line in f:
            line = line.rstrip()
            known_svs.append(line.split('\t')[0:3])
            # print(line.split('\t')[0:3])
    # returns a list of known svs as a list where each element is [chr_name, st, end]
    return known_svs


def count_miassemblies_in_autosomes(misassemblies):
    autosomes = ['chr' + str(x) for x in range(1, 23)] + ['chrX', 'chrY']
    total_count = 0
    relocation_count = 0
    translocation_count = 0
    inversion_count = 0

    for chr, st, end, ms_type in misassemblies:
        if chr in autosomes:
            if ms_type == 'relocation':
                relocation_count += 1
                total_count += 1
            elif ms_type == 'translocation':
                translocation_count += 1
                total_count += 1
            elif ms_type == 'inversion':
                inversion_count += 1
                total_count += 1
        else:
            print(chr, st, end, ms_type)

    print("#####--TOTAL MISSASSEMBLIES: IN AUTOSOMES--#####")
    print("Total Misassemblies:\t", total_count)
    print("Total relocations:\t", relocation_count)
    print("Total translocations:\t", translocation_count)
    print("Total inversions:\t", inversion_count)
    print("################################################\n")


def count_misassemblies_not_overlapping_with_svs(known_svs, misassemblies):
    total_count = 0
    total_base_count = 0
    relocation_count = 0
    translocation_count = 0
    inversion_count = 0
    autosomes = ['chr' + str(x) for x in range(1, 23)] + ['chrX', 'chrY']

    for chr_ms, st_ms, end_ms, ms_type in misassemblies:
        if chr_ms not in autosomes:
            continue

        overlaps = False
        for chr_sv, st_sv, end_sv in known_svs:
            # x1 <= y2 && y1 <= x2
            if chr_ms == chr_sv and int(st_ms) <= int(end_sv) and int(st_sv) <= int(end_ms):
                # print(chr_ms, st_ms, end_ms, ms_type, 'overlaps with sv', chr_sv, st_sv, end_sv)
                overlaps = True
                break
            if overlaps:
                break

        if overlaps is False:
            total_count += 1
            total_base_count += (int(end_ms) - int(st_ms) + 1)
            if ms_type == 'relocation':
                relocation_count += 1
            elif ms_type == 'translocation':
                translocation_count += 1
            elif ms_type == 'inversion':
                inversion_count += 1

    print("#####--MISSASSEMBLIES OUTSIDE SVs--#####")
    print("Total Misassemblies:\t", total_count)
    print("Total relocations:\t", relocation_count)
    print("Total translocations:\t", translocation_count)
    print("Total inversions:\t", inversion_count)
    print("Total Bases:\t", total_base_count/ 1000000000)
    print("########################################")


if __name__ == '__main__':
    '''
    Processes arguments and performs tasks.
    '''
    parser = argparse.ArgumentParser(description="Reports QUAST misassemblies that happen outside the known SVs.")
    parser.add_argument(
        "-q",
        "--quast_file",
        type=str,
        required=True,
        help="Path to the QUAST's TSV input file located in /QUAST_output/contig_reports/all_alignments_$$.tsv"
    )
    parser.add_argument(
        "-s",
        "--sv_file",
        type=str,
        required=True,
        help="Path to a bed file containing all known SVs for the sample."
    )
    parser.add_argument(
        "-c",
        "--cen_file",
        type=str,
        required=True,
        help="Path to a bed file containing all known centromeric region."
    )
    parser.add_argument(
        "-d",
        "--segdup_file",
        type=str,
        required=True,
        help="Path to a bed file containing all known segdup for the sample."
    )
    FLAGS, unparsed = parser.parse_known_args()
    # get regions from files
    svs = read_bed_file(FLAGS.sv_file)
    centromeres = read_bed_file(FLAGS.cen_file)
    segdups = read_bed_file(FLAGS.segdup_file)

    centromeres_only = centromeres
    centromeres_and_segdups = centromeres + segdups
    m_assemblies = read_quast_file(FLAGS.quast_file)

    # count stuff
    # count_miassemblies_in_autosomes(m_assemblies)
    print("OUTSIDE CENTROMERES")
    count_misassemblies_not_overlapping_with_svs(centromeres_only, m_assemblies)
    print("\nOUTSIDE CENTROMERES AND SEG DUPS")
    count_misassemblies_not_overlapping_with_svs(centromeres_and_segdups, m_assemblies)


import h5py
import sys
import concurrent.futures
import numpy as np
from collections import defaultdict
import operator
from modules.python.TextColor import TextColor
from modules.python.Options import StitchOptions
from modules.python.FileManager import FileManager
from build import HELEN
import re


class Stitch:
    """
    This class performs stitching of sequences predicted by HELEN in chunks.

    The method is as follows:
    1) MarginPolish
     - MarginPolish takes a "Region" which is usually 1000 bases and generates pileup summaries for them.
     - The region often is more than 1000 bases but for mini-batch to work, we make sure all the images are 1000 bases
       so MarginPolish chunks these Regions into images providing each Regional image with a chunk id.
    2) HELEN call consensus
     - Call Consensus generates a prediction for each of the images, but while saving it makes sure all the images
       that belong to the same region, gets saved under the same chunk_prefix, making sure all the sequences can
     - be aggregated easily.
    3) HELEN stitch
     -  The stitch method that loads one contig at a time and grabs all the "regions" with predictions. The regional
        chunks are easily stitched without having to do an overlap as the positions inside a region is consistent.
     - For regions that are adjacent, HELEN uses a local smith-waterman alignment to find an anchor point to
       stitch sequences from two adjacent sequences together.
    """
    @staticmethod
    def get_confident_positions(alignment):
        """
        Given an alignment object, this method returns a position in both sequences that we can use as
        anchor positions. This method looks for a significant overlap between two sequences that we can use
        as an anchor.
        :param alignment: SSW alignment object.
        :return: two positions
        """
        # get the cigar string and replace '=' and 'X' with 'M' as both can represent a base alignment.
        cigar_string = alignment.cigar_string.replace('=', 'M').replace('X', 'M')

        # this regex will split the cigar string into a list of tuples where each item is (cigar_operation, cigar_len)
        cigar_tuples = re.findall(r'(\d+)(\w)', cigar_string)

        # now we would group the 'M's together which we derived from '=' and 'X'.
        grouped_tuples = list()
        prev_len = 0
        prev_op = None
        for cigar_len, cigar_op in cigar_tuples:
            # if there were no previous operations
            if prev_op is None:
                prev_op = cigar_op
                prev_len = int(cigar_len)
            # if the previous operation is the same as current operation
            elif prev_op == cigar_op:
                # simply extend the operation
                prev_len += int(cigar_len)
            else:
                grouped_tuples.append((prev_op, prev_len))
                prev_op = cigar_op
                prev_len = int(cigar_len)

        # the last one that won't get appeneded
        if prev_op is not None:
            grouped_tuples.append((prev_op, prev_len))

        ref_index = alignment.reference_begin
        read_index = 0

        # now look for an anchor position. We define an anchor position to be a position where we've seen atleast
        # 5 bases overlap between two sequences.
        for cigar_op, cigar_len in grouped_tuples:
            if cigar_op == 'M' and cigar_len >= StitchOptions.OVERLAP_THRESHOLD:
                return ref_index, read_index

            if cigar_op == 'S':
                read_index += cigar_len
            elif cigar_op == 'I':
                read_index += cigar_len
            elif cigar_op == 'D':
                ref_index += cigar_len
            elif cigar_op == 'M':
                ref_index += cigar_len
                read_index += cigar_len
            else:
                # this has never happened so far.
                raise ValueError(TextColor.RED + "ERROR: INVALID CIGAR OPERATION ENCOUNTERED WHILTE STITCHING: "
                                 + str(cigar_op) + "\n")

        # if we can't find any anchors return negative values
        return -1, -1

    def alignment_stitch(self, sequence_chunks):
        """
        This is a stitch worker. This method gets a chunk of contiguous sequence that it stitches.
        The method is very simple, it performs an ssw alignment, finds an anchor position and concatenates
        two adjacent sequences that overlap.
        :param sequence_chunks: A list of sequence chunks in (contig, start, end, sequence) format.
        :return:
        """
        # we make sure that the chunks are sorted by the positions
        sequence_chunks = sorted(sequence_chunks, key=lambda element: (element[1], element[2]))
        # pick the first sequence to be the running sequence
        contig, running_start, running_end, running_sequence = sequence_chunks[0]

        # initialize an ssw aligner
        aligner = HELEN.Aligner(StitchOptions.MATCH_PENALTY, StitchOptions.MISMATCH_PENALTY,
                                StitchOptions.GAP_PENALTY, StitchOptions.GAP_EXTEND_PENALTY)
        # and a filter required by ssw align function
        filter = HELEN.Filter()

        # now iterate through all the chunks
        for i in range(1, len(sequence_chunks)):
            # get the current suquence
            _, this_start, this_end, this_sequence = sequence_chunks[i]
            # make sure the current sequence overlaps with the previously processed sequence
            if this_start < running_end:
                # overlap
                overlap_bases = running_end - this_start
                overlap_bases = overlap_bases + int(overlap_bases * StitchOptions.BASE_ERROR_RATE)

                # now we take the last bases from running sequence
                left_running_sequence_chunk = running_sequence[-overlap_bases:]
                # and first bases from the current sequence
                right_current_sequence = this_sequence[:overlap_bases]

                # initialize an alignment object
                alignment = HELEN.Alignment()
                aligner.SetReferenceSequence(left_running_sequence_chunk, len(left_running_sequence_chunk))
                # align current sequence to the previous sequence
                aligner.Align_cpp(right_current_sequence, filter, alignment, 0)

                # check we have an alignment between the sequences
                if alignment.best_score == 0:
                    sys.stderr.write(TextColor.YELLOW + "WARNING: NO ALIGNMENT FOUND: " + str(this_start)
                                     + " " + str(this_end) + "\n" + TextColor.END)
                    # this is a special case, happens when we encounter a region that is empty. In this case what we do
                    # is append 50 Ns to compensate for the overlap regions and then add the next chunk. This happens
                    # very rarely but happens for sure.
                    if len(right_current_sequence) > 10:
                        running_sequence = running_sequence + 10 * 'N'
                        running_sequence = running_sequence + right_current_sequence
                        running_end = this_end

                else:
                    # we have a valid aignment so we try to find an anchor position
                    pos_a, pos_b = self.get_confident_positions(alignment)

                    if pos_a == -1 or pos_b == -1:
                        # in this case we couldn't find a place that we can use as an anchor
                        # we again compensate this Ns in the sequence.
                        sys.stderr.write(TextColor.YELLOW + "WARNING: NO OVERLAPS IN ALIGNMENT : \n" + TextColor.END)
                        sys.stderr.write(TextColor.YELLOW + "LEFT : " + str(left_running_sequence_chunk) + "\n" +
                                         TextColor.END)
                        sys.stderr.write(TextColor.YELLOW + "RIGHT: " + str(right_current_sequence) + "\n" +
                                         TextColor.END)
                        sys.stderr.write(TextColor.YELLOW + "CIGAR: " + str(alignment.cigar_string) + "\n" +
                                         TextColor.END)
                        if len(this_sequence) > 10:
                            left_sequence = running_sequence[:-overlap_bases]
                            overlap_sequence = left_running_sequence_chunk
                            running_sequence = left_sequence + overlap_sequence + 10 * 'N' + this_sequence
                            running_end = this_end
                    else:
                        # this is a perfect match so we can simply stitch them
                        # take all of the sequence from the left
                        left_sequence = running_sequence[:-overlap_bases]
                        # get the bases that overlapped
                        overlap_sequence = left_running_sequence_chunk[:pos_a]
                        # get sequences from current sequence
                        right_sequence = this_sequence[pos_b:]

                        # now append all three parts and we have a contiguous sequence
                        running_sequence = left_sequence + overlap_sequence + right_sequence
                        running_end = this_end
            else:
                # in this case we encountered a region where there's no high level overlap.
                # In this case we again compensate with Ns.
                sys.stderr.write(TextColor.YELLOW + "WARNING: NO OVERLAP IN CHUNKS: " + " " + str(contig)
                                 + " " + str(this_start) + " " + str(running_end) + "\n" + TextColor.END)

                # if the sequence is worth adding, then we add
                if len(this_sequence) > 10:
                    running_sequence = running_sequence + 10 * 'N' + this_sequence
                    running_end = this_end

        return contig, running_start, running_end, running_sequence

    def small_chunk_stitch(self, contig, small_chunk_keys):
        """
        This process stitches regional image chunk predictions. Among these image chunks, the positions are
        always consistent, we don't need an alignment to stitch these chunks.
        :param contig: Contig name
        :param small_chunk_keys: Chunk keys in list as (contig_name, start_position, end_position)
        :return:
        """
        # for chunk_key in small_chunk_keys:
        name_sequence_tuples = list()

        # go through all the chunk keys
        for contig_name, file_name, chunk_name, contig_start, contig_end in small_chunk_keys:
            smaller_chunks = []
            with h5py.File(file_name, 'r') as hdf5_file:
                if 'predictions' in hdf5_file:
                    smaller_chunks = set(hdf5_file['predictions'][contig][chunk_name].keys()) - \
                                     {'contig_start', 'contig_end'}

            smaller_chunks = sorted(smaller_chunks)
            all_positions = set()
            # now create two dictionaries where we will save all the predictions
            base_prediction_dict = defaultdict()
            rle_prediction_dict = defaultdict()
            for chunk in smaller_chunks:
                # grab the predictions and the positions
                with h5py.File(file_name, 'r') as hdf5_file:
                    bases = hdf5_file['predictions'][contig][chunk_name][chunk]['bases'][()]
                    rles = hdf5_file['predictions'][contig][chunk_name][chunk]['rles'][()]
                    positions = hdf5_file['predictions'][contig][chunk_name][chunk]['position'][()]

                positions = np.array(positions, dtype=np.int64)
                base_predictions = np.array(bases, dtype=np.int)
                rle_predictions = np.array(rles, dtype=np.int)

                # now iterate over each position and add the predictions to the dictionary
                for position, base_pred, rle_pred in zip(positions, base_predictions, rle_predictions):
                    indx = position[1]
                    pos = position[0]
                    split_indx = position[2]
                    if indx < 0 or pos < 0:
                        continue
                    if (pos, indx, split_indx) not in base_prediction_dict:
                        base_prediction_dict[(pos, indx, split_indx)] = base_pred
                        rle_prediction_dict[(pos, indx, split_indx)] = rle_pred
                        all_positions.add((pos, indx, split_indx))

            # now simply create a position list and query the  dictionary to generate the predicted sequence
            pos_list = sorted(list(all_positions), key=lambda element: (element[0], element[1], element[2]))
            dict_fetch = operator.itemgetter(*pos_list)
            predicted_base_labels = list(dict_fetch(base_prediction_dict))
            predicted_rle_labels = list(dict_fetch(rle_prediction_dict))
            sequence = ''.join([StitchOptions.label_decoder[base] * int(rle)
                                for base, rle in zip(predicted_base_labels, predicted_rle_labels)])
            # now add the generated sequence for further stitching
            name_sequence_tuples.append((contig, contig_start, contig_end, sequence))

        # now we have all the regional sequences generated we can add them using ssw.
        name_sequence_tuples = sorted(name_sequence_tuples, key=lambda element: (element[1], element[2]))

        # stich using ssw and return the contiguous sequence from the regional chunks we got from here.
        contig, running_start, running_end, running_sequence = self.alignment_stitch(name_sequence_tuples)

        return contig, running_start, running_end, running_sequence

    def create_consensus_sequence(self, contig, sequence_chunk_keys, threads):
        """
        This is the consensus sequence create method that creates a sequence for a given contig.
        :param contig: Contig name
        :param sequence_chunk_keys: All the chunk keys in the contig
        :param threads: Number of available threads
        :return: A consensus sequence for a contig
        """
        # first we sort the sequence chunks
        sequence_chunk_key_list = list()

        # then we split the chunks to so get contig name, start and end positions so we can sort them properly
        for hdf5_file, chunk_key, st, end in sequence_chunk_keys:
            sequence_chunk_key_list.append((contig, hdf5_file, chunk_key, int(st), int(end)))

        # we sort based on positions
        sequence_chunk_key_list = sorted(sequence_chunk_key_list, key=lambda element: (element[3], element[4]))

        sequence_chunks = list()
        # we submit the chunks in process pool
        with concurrent.futures.ProcessPoolExecutor(max_workers=threads) as executor:
            # this chunks the keys into sucessive chunks
            file_chunks = FileManager.chunks(sequence_chunk_key_list,
                                             max(StitchOptions.MIN_SEQUENCE_REQUIRED_FOR_MULTITHREADING,
                                                 int(len(sequence_chunk_key_list) / threads) + 1))

            # we do the stitching per chunk of keys
            futures = [executor.submit(self.small_chunk_stitch, contig, file_chunk)
                       for file_chunk in file_chunks]

            # as they complete we add them to a list
            for fut in concurrent.futures.as_completed(futures):
                if fut.exception() is None:
                    contig, contig_start, contig_end, sequence = fut.result()
                    sequence_chunks.append((contig, contig_start, contig_end, sequence))
                else:
                    sys.stderr.write("ERROR: " + str(fut.exception()) + "\n")
                fut._result = None  # python issue 27144

        sequence_chunks = sorted(sequence_chunks, key=lambda element: (element[1], element[2]))

        # and do a final stitching on all the sequences we generated
        contig, contig_start, contig_end, sequence = self.alignment_stitch(sequence_chunks)

        return sequence



import os
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel
import sys
import torch
import torch.nn as nn
import time
from torch.utils.data import DataLoader
from modules.python.models.dataloader_predict import SequenceDataset
from modules.python.TextColor import TextColor
from modules.python.models.ModelHander import ModelHandler
from modules.python.Options import ImageSizeOptions, TrainOptions
from modules.python.DataStore import DataStore
"""
This script implements the predict method that is used by the call consensus method.

The algorithm is described here:

  1) INPUTS:
    - directory path to the image files generated by MarginPolish
    - model path directing to a trained model
    - batch size for minibatch prediction
    - num workers for minibatch processing threads
    - output directory path to where the output hdf5 will be saved
  2) METHOD:
    - Call predict function that loads the neural network and generates base predictions and saves it into a hdf5 file
        - Loads the model
        - Iterates over the input images in minibatch
        - For each image uses a sliding window method to slide of the image sequence
        - Aggregate the predictions to get sequence prediction for the entire image sequence
        - Save all the predictions to a file
  3) OUTPUT:
    - A hdf5 file containing all the base predictions   
"""


def predict(test_file, output_filename, model_path, batch_size, num_workers, rank, device_id):
    """
    The predict method loads images generated by MarginPolish and produces base predictions using a
    sequence transduction model based deep neural network. This method loads the model and iterates over
    minibatch images to generate the predictions and saves the predictions to a hdf5 file.

    :param test_file: File to predict on
    :param output_filename: Name and path to the output file
    :param batch_size: Batch size used for minibatch prediction
    :param model_path: Path to a trained model
    :param rank: Rank of this caller
    :param num_workers: Number of workers to be used by the dataloader
    :param device_id: Device id of the model to set to
    :param threads: Number of threads to use with pytorch
    :return: Prediction dictionary
    """
    # create the output hdf5 file where all the predictions will be saved
    prediction_data_file = DataStore(output_filename + "_" + str(rank) + ".hdf", mode='w')

    # load the model using the model path
    transducer_model, hidden_size, gru_layers, prev_ite = \
        ModelHandler.load_simple_model(model_path,
                                       input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                       image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                       seq_len=ImageSizeOptions.SEQ_LENGTH,
                                       num_base_classes=ImageSizeOptions.TOTAL_BASE_LABELS,
                                       num_rle_classes=ImageSizeOptions.TOTAL_RLE_LABELS)
    transducer_model.eval()

    torch.cuda.set_device(device_id)
    transducer_model.to(device_id)
    transducer_model.eval()
    transducer_model = DistributedDataParallel(transducer_model, device_ids=[device_id])

    # only output for rank 0 caller
    if rank == 0:
        print(output_filename + "_" + str(rank) + ".hdf")
        sys.stderr.write(TextColor.GREEN + 'INFO: TORCH THREADS SET TO: ' + str(torch.get_num_threads()) + ".\n"
                         + TextColor.END)
        # notify that the process has started and loading data
        sys.stderr.write(TextColor.PURPLE + 'Loading data\n' + TextColor.END)

    # create a pytorch dataset and dataloader that loads the data in mini_batches
    test_data = SequenceDataset(image_directory=None, file_list=test_file)
    test_loader = DataLoader(test_data,
                             batch_size=batch_size,
                             shuffle=False,
                             num_workers=num_workers)

    # iterate over the data in minibatches
    with torch.no_grad():
        # keep an eye for batch
        total_batches = len(test_loader)
        batch_iterator = 0

        # the dataloader loop, iterates in minibatches. tqdm is the progress logger.
        for contig, contig_start, contig_end, chunk_id, images, position, filename in test_loader:
            start_time = time.time()
            # the images are usually in uint8, convert them to FloatTensor
            images = images.type(torch.FloatTensor)
            # initialize the first hidden input as all zeros
            hidden = torch.zeros(images.size(0), 2 * TrainOptions.GRU_LAYERS, TrainOptions.HIDDEN_SIZE)

            # this is a multi-task neural network where we predict a base and a run-length. We use two dictionaries
            # to keep track of predictions.
            # these two dictionaries save predictions for each of the chunks and later we aggregate all the predictions
            # over the entire sequence to get a sequence prediction for the whole sequence.
            prediction_base_tensor = torch.zeros((images.size(0), images.size(1), ImageSizeOptions.TOTAL_BASE_LABELS))
            prediction_rle_tensor = torch.zeros((images.size(0), images.size(1), ImageSizeOptions.TOTAL_RLE_LABELS))

            # move the tensors to CUDA
            prediction_base_tensor = prediction_base_tensor.to(device_id)
            prediction_rle_tensor = prediction_rle_tensor.to(device_id)

            # now the images usually contain 1000 bases, we iterate on a sliding window basis where we process
            # the window size then jump to the next window
            for i in range(0, ImageSizeOptions.SEQ_LENGTH, TrainOptions.WINDOW_JUMP):
                # if current position + window size goes beyond the size of the window, that means we've reached the end
                if i + TrainOptions.TRAIN_WINDOW > ImageSizeOptions.SEQ_LENGTH:
                    break
                chunk_start = i
                chunk_end = i + TrainOptions.TRAIN_WINDOW

                # get the image chunk
                image_chunk = images[:, chunk_start:chunk_end]

                # move the images and the hidden layers to CUDA
                images = images.to(device_id)
                hidden = hidden.to(device_id)

                # run inference
                output_base, output_rle, hidden = transducer_model(image_chunk, hidden)

                # now calculate how much padding is on the top and bottom of this chunk so we can do a simple
                # add operation
                top_zeros = chunk_start
                bottom_zeros = ImageSizeOptions.SEQ_LENGTH - chunk_end

                # we run a softmax a padding to make the output tensor compatible for adding
                inference_layers = nn.Sequential(
                    nn.Softmax(dim=2),
                    nn.ZeroPad2d((0, 0, top_zeros, bottom_zeros))
                )
                inference_layers = inference_layers.to(device_id)

                # run the softmax and padding layers
                base_prediction = inference_layers(output_base)
                rle_prediction = inference_layers(output_rle)

                # now simply add the tensor to the global counter
                prediction_base_tensor = torch.add(prediction_base_tensor, base_prediction)
                prediction_rle_tensor = torch.add(prediction_rle_tensor, rle_prediction)

            # all done now create a SEQ_LENGTH long prediction list
            prediction_base_tensor = prediction_base_tensor.cpu()
            prediction_rle_tensor = prediction_rle_tensor.cpu()

            base_values, base_labels = torch.max(prediction_base_tensor, 2)
            rle_values, rle_labels = torch.max(prediction_rle_tensor, 2)

            predicted_base_labels = base_labels.cpu().numpy()
            predicted_rle_labels = rle_labels.cpu().numpy()

            batch_iterator += 1

            if rank == 0:
                # calculate the expected time to finish
                eta = (time.time() - start_time) * (total_batches - batch_iterator)
                hours = str(int(eta/3600))
                eta = eta - (eta/3600)
                mins = str(int(eta/60))
                secs = str(int(eta) % 60)
                time_stamp = hours + " HOURS " + mins + " MINS " + secs + " SECS."
                batch_string = "BATCHES DONE: " + str(batch_iterator) + "/" + str(total_batches) + ". "
                time_left = "ESTIMATED TIME LEFT: " + str(time_stamp)
                sys.stderr.write(TextColor.GREEN + "INFO: " + batch_string + time_left + "\n" + TextColor.END)

            # go to each of the images and save the predictions to the file
            for i in range(images.size(0)):
                prediction_data_file.write_prediction(contig[i], contig_start[i], contig_end[i], chunk_id[i],
                                                      position[i], predicted_base_labels[i], predicted_rle_labels[i],
                                                      filename[i])


def cleanup():
    dist.destroy_process_group()


def setup(rank, total_callers, args, all_input_files, all_devices):
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'

    # initialize the process group
    dist.init_process_group("gloo", rank=rank, world_size=total_callers)

    # expand the arguments
    output_filepath, model_path, batch_size, num_workers = args

    # call prediction function
    predict(all_input_files[rank],
            output_filepath,
            model_path,
            batch_size,
            num_workers,
            rank,
            all_devices[rank])
    cleanup()


def predict_gpu(file_chunks, output_filepath, model_path, batch_size, total_callers, devices, num_workers):
    """
    Create a prediction table/dictionary of an images set using a trained model.
    :param file_chunks: Path to chunked files
    :param batch_size: Batch size used for prediction
    :param model_path: Path to a trained model
    :param output_filepath: Path to output directory
    :param total_callers: Number of callers to spawn
    :param devices: List of available CUDA devices
    :param num_workers: Number of workers to be used by the dataloader
    :return: Prediction dictionary
    """
    # create the arguments to send for prediction
    args = (output_filepath, model_path, batch_size, num_workers)

    # spawn the processes to call the prediction method
    mp.spawn(setup,
             args=(total_callers, args, file_chunks, devices),
             nprocs=total_callers,
             join=True)



import pandas as pd
from collections import defaultdict

medaka_table = pd.read_csv('/data/users/kishwar/marginpolish_output/bac_analysis/medaka_hg00733_bac_stats.txt', sep='\t')


def is_overlapping(x1,x2,y1,y2):
    return max(x1,y1) <= min(x2,y2)


alignment_dict = defaultdict()
alignment_intervals = []
count = 0
for index, row in medaka_table.iterrows():
    if float(row['coverage']) == 100.0:
        count += 1
        alignment_dict[str(row['name'])] = True
        alignment_intervals.append((str(row['name']), str(row['ref']), int(row['rstart']), int(row['rend'])))


for i in range(len(alignment_intervals)):
    for j in range(len(alignment_intervals)):
        if i == j:
            continue
        if alignment_intervals[i][1] != alignment_intervals[j][1]:
            continue
        if is_overlapping(alignment_intervals[i][2], alignment_intervals[j][2],
                          alignment_intervals[i][3], alignment_intervals[j][3]):
            alignment_dict[alignment_intervals[i][0]] = False
            alignment_dict[alignment_intervals[j][0]] = False

for i, alignment_name in enumerate(alignment_dict.keys()):
    if alignment_dict[alignment_name] is True:
        print(i, alignment_name)
print(count)



import os
import re
import sys
import shutil

from setuptools import setup, Extension
from setuptools import Distribution
from setuptools.command.install import install
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion
from glob import glob
import subprocess

__pkg_name__ = 'helen'
__author__ = 'Kishwar Shafin'
__description__ = 'RNN based assembly HELEN. It works paired with MarginPolish.'


class CMakeExtension(Extension):
    def __init__(self, name):
        Extension.__init__(self, name, sources=[])


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)',
                                                   out.decode()).group(1))
            print("CMAKE VERSION: ", cmake_version)
            if cmake_version < '3':
                raise RuntimeError("CMake >= 3 is required.")
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: " +
                ", ".join(e.name for e in self.extensions))

        build_directory = os.path.abspath(self.build_temp)

        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + build_directory,
            '-DPYTHON_EXECUTABLE=' + sys.executable
        ]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]

        # Assuming Makefiles
        build_args += ['--', '-j2']

        self.build_args = build_args

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get('CXXFLAGS', ''),
            self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)

        # CMakeLists.txt is in the same directory as this setup.py file
        cmake_list_dir = os.path.abspath(os.path.dirname(__file__))
        print('-'*10, 'Running CMake prepare', '-'*40)
        subprocess.check_call(['cmake', cmake_list_dir] + cmake_args,
                              cwd=self.build_temp, env=env)

        print('-'*10, 'Building extensions', '-'*40)
        cmake_cmd = ['cmake', '--build', '.'] + self.build_args
        subprocess.check_call(cmake_cmd,
                              cwd=self.build_temp)

        # Move from build temp to final position
        for ext in self.extensions:
            self.move_output(ext)

    def move_output(self, ext):
        source_path = os.path.abspath(self.build_temp + "/" + self.get_ext_filename(ext.name))

        dest_directory = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name))) + "/build/"
        os.makedirs(dest_directory, exist_ok=True)

        dest_path = dest_directory + self.get_ext_filename(ext.name)
        self.copy_file(source_path, dest_path)

        # move marginpolish source
        mp_source_path = os.path.abspath(self.build_temp + "/" + "marginpolish/src/marginpolish-build/marginPolish")

        mp_dest_directory = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name))) + "/bin"

        os.makedirs(mp_dest_directory, exist_ok=True)

        mp_dest_path = mp_dest_directory + "/marginPolish"
        self.copy_file(mp_source_path, mp_dest_path)


def get_dependencies():
    # create requirements from requirements.txt
    dir_path = os.path.dirname(__file__)
    install_requires = []
    with open(os.path.join(dir_path, 'requirements.txt')) as fh:
        reqs = (
            r.split('#')[0].strip()
            for r in fh.read().splitlines() if not r.strip().startswith('#')
        )
        for req in reqs:
            if req.startswith('git+https'):
                req = req.split('/')[-1].split('@')[0]
            install_requires.append(req)
    return install_requires


def get_version():
    version = {}
    with open("version.py") as fp:
        exec(fp.read(), version)
    return version['__version__']


def get_long_description():
    from os import path
    this_directory = path.abspath(path.dirname(__file__))
    kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
    with open(path.join(this_directory, 'README.md'), **kwargs) as f:
        long_description = f.read()
    long_description_content_type = 'text/markdown'

    return long_description, long_description_content_type


# Nasty hack to get binaries into bin path
class GetPaths(install):
    def run(self):
        self.distribution.install_scripts = self.install_scripts
        self.distribution.install_libbase = self.install_libbase


def get_setuptools_marginpolish_dir():
    dist = Distribution({'cmdclass': {'install': GetPaths}})
    dist.dry_run = True
    dist.parse_config_files()
    command = dist.get_command_obj('install')
    command.ensure_finalized()
    command.run()

    src_dir = glob(os.path.join(dist.install_libbase, 'helen-*', 'bin'))[0]

    for exe in (os.path.join(src_dir, x) for x in os.listdir(src_dir)):
        shutil.copy(exe, dist.install_scripts)
    return dist.install_libbase, dist.install_scripts


if __name__ == '__main__':
    # check python3 version
    pymajor, pyminor = sys.version_info[0:2]
    if (pymajor < 3) or (pymajor <= 3 and pyminor < 5):
        raise RuntimeError(
            'HELEN requires 3.5 higher.')
    python_dependencies = get_dependencies()
    __long_description__, __long_description_content_type__ = get_long_description()

    setup(
        name='helen',
        version=get_version(),
        packages=['', 'modules/python', 'modules/python/models', 'modules/python/helper'],
        package_dir={'modules/python': 'modules/python',
                     'modules/python/models': 'modules/python/models',
                     'modules/python/helper': 'modules/python/helper'},
        url='https://github.com/kishwarshafin/helen',
        author=__author__,
        author_email="kishwar.shafin@gmail.com",
        description=__description__,
        long_description=__long_description__,
        long_description_content_type=__long_description_content_type__,
        python_requires='>=3.5.*',
        install_requires=python_dependencies,
        entry_points={
            'console_scripts': [
                '{0} = {0}:main'.format(__pkg_name__),
                '{0}_train = {0}_train:main'.format(__pkg_name__)
            ]
        },
        ext_modules=[CMakeExtension('HELEN')],
        cmdclass={
            'build_ext': CMakeBuild
        },
        zip_safe=False,
    )

    if os.environ.get('INSTALL_BINARIES') is not None:
        print("\nCopying utility binaries to your path.")
        get_setuptools_marginpolish_dir()



import pandas as pd
import os
import sys
import h5py
from tqdm import tqdm

csv_path = sys.argv[1]

tmp_df = pd.read_csv(csv_path, header=None)
img_files = list(tmp_df[0])
allele_path = list(tmp_df[2])
for i in tqdm(range(len(img_files))):
    hdf5_file_path = img_files[i]
    allele_dict_path = allele_path[i]
    if os.path.isfile(hdf5_file_path) is False:
        print("INVALID FILE PATH: ", hdf5_file_path)
        exit()
    elif os.path.isfile(allele_dict_path) is False:
        print("INVALID FILE PATH: ", allele_dict_path)
        exit()
    else:
        hdf5_file = h5py.File(hdf5_file_path, 'r')
        if 'images' not in hdf5_file.keys():
            print("NO IMAGES IN HDF5", hdf5_file_path)
            exit()



import torch
import os
from modules.python.models.TransducerModel import TransducerGRU


class ModelHandler:
    """
    The ModelHandler class handles different model saving/loading operations that we do.
    """
    @staticmethod
    def save_checkpoint(state, filename):
        """
        Save a model checkpoint to a file.
        :param state: A model state (usually the model itself in a dictionary format)
        :param filename: Name of the output file
        :return:
        """
        torch.save(state, filename)

    @staticmethod
    def get_new_gru_model(input_channels, image_features, gru_layers, hidden_size, num_base_classes, num_rle_classes):
        """
        Create a new model object and return it.
        :param input_channels: Number of channels in the input image (usually 1)
        :param image_features: Number of features in one column of the pileup
        :param gru_layers: Number of layers in the transducer model
        :param hidden_size: The size of the hidden layer
        :param num_base_classes: Number of base classes
        :param num_rle_classes: Number of RLE classes
        :return: A new model object
        """
        # get a new model
        transducer_model = TransducerGRU(input_channels, image_features, gru_layers, hidden_size, num_base_classes,
                                         num_rle_classes, bidirectional=True)
        return transducer_model

    @staticmethod
    def load_simple_model(model_path, input_channels, image_features, seq_len, num_base_classes, num_rle_classes):
        """
        This method loads a model from a given model path.
        :param model_path: Path to a model
        :param input_channels: Number of channels in the input image (usually 1)
        :param image_features: Number of features in one column of the pileup
        :param seq_len: Length of the sequence in one image
        :param num_base_classes: Number of base classes
        :param num_rle_classes: Number of RLE classes
        :return: A loaded model with some other auxiliary information
        """
        # first load the model to cpu, it's usually a dicttionary
        checkpoint = torch.load(model_path, map_location='cpu')
        # extract auxiliary information from the model dictionary
        hidden_size = checkpoint['hidden_size']
        gru_layers = checkpoint['gru_layers']
        epochs = checkpoint['epochs']

        # create a new model
        transducer_model = ModelHandler.get_new_gru_model(input_channels=input_channels,
                                                          image_features=image_features,
                                                          gru_layers=gru_layers,
                                                          hidden_size=hidden_size,
                                                          num_base_classes=num_base_classes,
                                                          num_rle_classes=num_rle_classes)
        # load the model state/weights
        model_state_dict = checkpoint['model_state_dict']

        # now create a state dict that we can load to the new model
        from collections import OrderedDict
        new_model_state_dict = OrderedDict()

        for k, v in model_state_dict.items():
            name = k
            # this happens due to training on the GPU. It's a pytorch issue, we can't fix it.
            if k[0:7] == 'module.':
                name = k[7:]  # remove `module.`
            new_model_state_dict[name] = v

        # transfer the weights to the new model
        transducer_model.load_state_dict(new_model_state_dict)
        transducer_model.cpu()

        # return the loaded model
        return transducer_model, hidden_size, gru_layers, epochs

    @staticmethod
    def load_simple_optimizer(transducer_optimizer, checkpoint_path, gpu_mode):
        """
        Load the optimizer state. This is required during re-training/transfer learning a previous model.
        :param transducer_optimizer: Optimizer state
        :param checkpoint_path: Path to the model
        :param gpu_mode: If True, the model and optimizer will be loaded on CUDA/GPU.
        :return: A loaded optimizer
        """
        if gpu_mode:
            # if gpu is true, then load the optimizer and transfer the state to cuda
            checkpoint = torch.load(checkpoint_path)
            transducer_optimizer.load_state_dict(checkpoint['model_optimizer'])
            for state in transducer_optimizer.state.values():
                for k, v in state.items():
                    if isinstance(v, torch.Tensor):
                        state[k] = v.cuda()
        else:
            # if gpu not true, then simply load the optimizer.
            checkpoint = torch.load(checkpoint_path, map_location='cpu')
            transducer_optimizer.load_state_dict(checkpoint['model_optimizer'])

        return transducer_optimizer

    @staticmethod
    def save_model(transducer_model,
                   model_optimizer,
                   hidden_size,
                   layers,
                   epoch,
                   file_name):
        """
        Save the model each epoch.
        :param transducer_model: Model for saving
        :param model_optimizer: Optimizer used for training the model
        :param hidden_size: Hidden layer size of the model
        :param layers: Number of layers in the model
        :param epoch: Number of epochs model has been trained on.
        :param file_name: Name of the file where the MODEL will be saved.
        :return:
        """
        if os.path.isfile(file_name):
            os.remove(file_name)
        ModelHandler.save_checkpoint({
            'model_state_dict': transducer_model.state_dict(),
            'model_optimizer': model_optimizer.state_dict(),
            'hidden_size': hidden_size,
            'gru_layers': layers,
            'epochs': epoch,
        }, file_name)





import sys
import torch
import os
from tqdm import tqdm
import torch.distributed as dist
import torch.nn as nn
import torch.multiprocessing as mp

from torch.utils.data import DataLoader
from modules.python.models.dataloader import SequenceDataset
from modules.python.TextColor import TextColor
from modules.python.models.ModelHander import ModelHandler
from modules.python.models.test import test
from modules.python.Options import ImageSizeOptions, TrainOptions

os.environ['PYTHONWARNINGS'] = 'ignore:semaphore_tracker:UserWarning'

"""
Train a model and return the model and optimizer trained.

Input:
- A train CSV containing training image set information (usually chr1-18)

Return:
- A trained model
"""
CLASS_WEIGHTS = [1.0, 1.0, 1.0, 1.0, 1.0]


def save_best_model(transducer_model, model_optimizer, hidden_size, layers, epoch,
                    file_name):
    """
    Save the best model
    :param transducer_model: A trained model
    :param model_optimizer: Model optimizer
    :param hidden_size: Number of hidden layers
    :param layers: Number of GRU layers to use
    :param epoch: Epoch/iteration number
    :param file_name: Output file name
    :return:
    """
    if os.path.isfile(file_name):
        os.remove(file_name)
    ModelHandler.save_checkpoint({
        'model_state_dict': transducer_model.state_dict(),
        'model_optimizer': model_optimizer.state_dict(),
        'hidden_size': hidden_size,
        'gru_layers': layers,
        'epochs': epoch,
    }, file_name)
    sys.stderr.write(TextColor.RED + "\nMODEL SAVED SUCCESSFULLY.\n" + TextColor.END)


def train(train_file, test_file, batch_size, epoch_limit, gpu_mode, num_workers, retrain_model,
          retrain_model_path, gru_layers, hidden_size, lr, decay, model_dir, stats_dir, train_mode,
          world_size, rank, device_id):

    if train_mode is True and rank == 0:
        train_loss_logger = open(stats_dir + "train_loss.csv", 'w')
        test_loss_logger = open(stats_dir + "test_loss.csv", 'w')
        confusion_matrix_logger = open(stats_dir + "confusion_matrix.txt", 'w')
    else:
        train_loss_logger = None
        test_loss_logger = None
        confusion_matrix_logger = None

    torch.cuda.set_device(device_id)

    if rank == 0:
        sys.stderr.write(TextColor.PURPLE + 'Loading data\n' + TextColor.END)

    train_data_set = SequenceDataset(train_file)

    train_sampler = torch.utils.data.distributed.DistributedSampler(
        train_data_set,
        num_replicas=world_size,
        rank=rank
    )

    train_loader = torch.utils.data.DataLoader(
        dataset=train_data_set,
        batch_size=batch_size,
        shuffle=False,
        num_workers=0,
        pin_memory=True,
        sampler=train_sampler)

    num_base_classes = ImageSizeOptions.TOTAL_BASE_LABELS
    num_rle_classes = ImageSizeOptions.TOTAL_RLE_LABELS

    if retrain_model is True:
        if os.path.isfile(retrain_model_path) is False:
            sys.stderr.write(TextColor.RED + "ERROR: INVALID PATH TO RETRAIN PATH MODEL --retrain_model_path\n")
            exit(1)
        sys.stderr.write(TextColor.GREEN + "INFO: RETRAIN MODEL LOADING\n" + TextColor.END)
        transducer_model, hidden_size, gru_layers, prev_ite = \
            ModelHandler.load_simple_model(retrain_model_path,
                                           input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                           image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                           seq_len=ImageSizeOptions.SEQ_LENGTH,
                                           num_base_classes=num_base_classes,
                                           num_rle_classes=num_rle_classes)

        if train_mode is True:
            epoch_limit = prev_ite + epoch_limit

        sys.stderr.write(TextColor.GREEN + "INFO: RETRAIN MODEL LOADED\n" + TextColor.END)
    else:
        transducer_model = ModelHandler.get_new_gru_model(input_channels=ImageSizeOptions.IMAGE_CHANNELS,
                                                          image_features=ImageSizeOptions.IMAGE_HEIGHT,
                                                          gru_layers=gru_layers,
                                                          hidden_size=hidden_size,
                                                          num_base_classes=num_base_classes,
                                                          num_rle_classes=num_rle_classes)
        prev_ite = 0

    param_count = sum(p.numel() for p in transducer_model.parameters() if p.requires_grad)
    if rank == 0:
        sys.stderr.write(TextColor.RED + "INFO: TOTAL TRAINABLE PARAMETERS:\t" + str(param_count) + "\n" + TextColor.END)

    model_optimizer = torch.optim.Adam(transducer_model.parameters(), lr=lr, weight_decay=decay)

    if retrain_model is True:
        sys.stderr.write(TextColor.GREEN + "INFO: OPTIMIZER LOADING\n" + TextColor.END)
        model_optimizer = ModelHandler.load_simple_optimizer(model_optimizer, retrain_model_path, gpu_mode)
        sys.stderr.write(TextColor.GREEN + "INFO: OPTIMIZER LOADED\n" + TextColor.END)

    if gpu_mode:
        transducer_model = transducer_model.to(device_id)
        transducer_model = nn.parallel.DistributedDataParallel(transducer_model, device_ids=[device_id])

    class_weights = torch.Tensor(TrainOptions.CLASS_WEIGHTS)
    # we perform a multi-task classification, so we need two loss functions, each performing a single task
    # criterion base is the loss function for base prediction
    criterion_base = nn.CrossEntropyLoss()
    # criterion rle is the loss function for RLE prediction
    criterion_rle = nn.CrossEntropyLoss(weight=class_weights)

    if gpu_mode is True:
        criterion_base = criterion_base.to(device_id)
        criterion_rle = criterion_rle.to(device_id)

    start_epoch = prev_ite

    # Train the Model
    if rank == 0:
        sys.stderr.write(TextColor.PURPLE + 'Training starting\n' + TextColor.END)
        sys.stderr.write(TextColor.BLUE + 'Start: ' + str(start_epoch + 1) + ' End: ' + str(epoch_limit) + "\n")

    stats = dict()
    stats['loss_epoch'] = []
    stats['accuracy_epoch'] = []

    for epoch in range(start_epoch, epoch_limit, 1):
        total_loss_base = 0
        total_loss_rle = 0
        total_loss = 0
        total_images = 0
        if rank == 0:
            sys.stderr.write(TextColor.BLUE + 'Train epoch: ' + str(epoch + 1) + "\n")
        # make sure the model is in train mode. BN is different in train and eval.

        batch_no = 1
        if rank == 0:
            progress_bar = tqdm(
                total=len(train_loader),
                ncols=100,
                leave=True,
                position=rank,
                desc="Loss: ",
            )
        else:
            progress_bar = None

        transducer_model.train()
        for images, label_base, label_rle in train_loader:
            # convert the tensors to the proper datatypes.
            images = images.type(torch.FloatTensor)
            label_base = label_base.type(torch.LongTensor)
            label_rle = label_rle.type(torch.LongTensor)

            hidden = torch.zeros(images.size(0), 2 * TrainOptions.GRU_LAYERS, TrainOptions.HIDDEN_SIZE)

            if gpu_mode:
                hidden = hidden.to(device_id)
                images = images.to(device_id)
                label_base = label_base.to(device_id)
                label_rle = label_rle.to(device_id)

            for i in range(0, ImageSizeOptions.SEQ_LENGTH, TrainOptions.WINDOW_JUMP):
                model_optimizer.zero_grad()

                if i + TrainOptions.TRAIN_WINDOW > ImageSizeOptions.SEQ_LENGTH:
                    break

                image_chunk = images[:, i:i+TrainOptions.TRAIN_WINDOW]
                label_base_chunk = label_base[:, i:i+TrainOptions.TRAIN_WINDOW]
                label_rle_chunk = label_rle[:, i:i+TrainOptions.TRAIN_WINDOW]

                # get the inference from the model
                output_base, output_rle, hidden = transducer_model(image_chunk, hidden)

                # calculate loss for base prediction
                loss_base = criterion_base(output_base.contiguous().view(-1, num_base_classes),
                                           label_base_chunk.contiguous().view(-1))
                # calculate loss for RLE prediction
                loss_rle = criterion_rle(output_rle.contiguous().view(-1, num_rle_classes),
                                         label_rle_chunk.contiguous().view(-1))

                # sum the losses to have a singlee optimization over multiple tasks
                loss = loss_base + loss_rle

                # backpropagation and weight update
                loss.backward()
                model_optimizer.step()

                # update the loss values
                total_loss += loss.item()
                total_loss_base += loss_base.item()
                total_loss_rle += loss_rle.item()
                total_images += image_chunk.size(0)

                # detach the hidden from the graph as the next chunk will be a new optimization
                hidden = hidden.detach()

            # update the progress bar
            avg_loss = (total_loss / total_images) if total_images else 0

            if train_mode is True and rank == 0:
                train_loss_logger.write(str(epoch + 1) + "," + str(batch_no) + "," + str(avg_loss) + "\n")

            if rank == 0:
                avg_loss = (total_loss / total_images) if total_images else 0
                progress_bar.set_description("Base: " + str(round(total_loss_base, 4)) +
                                             ", RLE: " + str(round(total_loss_rle, 4)) +
                                             ", TOTAL: " + str(round(total_loss, 4)))
                progress_bar.refresh()
                progress_bar.update(1)
                batch_no += 1

        if rank == 0:
            progress_bar.close()
        dist.barrier()

        if rank == 0:
            stats_dictionary = test(test_file, batch_size, gpu_mode, transducer_model, num_workers,
                                    gru_layers, hidden_size, num_base_classes=ImageSizeOptions.TOTAL_BASE_LABELS,
                                    num_rle_classes=ImageSizeOptions.TOTAL_RLE_LABELS)
            stats['loss'] = stats_dictionary['loss']
            stats['accuracy'] = stats_dictionary['accuracy']
            stats['loss_epoch'].append((epoch, stats_dictionary['loss']))
            stats['accuracy_epoch'].append((epoch, stats_dictionary['accuracy']))
        dist.barrier()

        # update the loggers
        if train_mode is True and rank == 0:
            ModelHandler.save_model(transducer_model, model_optimizer,
                                    hidden_size, gru_layers,
                                    epoch, model_dir + "HELEN_epoch_" + str(epoch + 1) + '_checkpoint.pkl')
            sys.stderr.write(TextColor.RED + "\nMODEL SAVED SUCCESSFULLY.\n" + TextColor.END)

            test_loss_logger.write(str(epoch + 1) + "," + str(stats['loss']) + "," + str(stats['accuracy']) + "\n")
            confusion_matrix_logger.write(str(epoch + 1) + "\n" + str(stats_dictionary['base_confusion_matrix']) + "\n")
            train_loss_logger.flush()
            test_loss_logger.flush()
            confusion_matrix_logger.flush()
        elif train_mode is False:
            # this setup is for hyperband
            if epoch + 1 >= 10 and stats['accuracy'] < 98:
                sys.stderr.write(TextColor.PURPLE + 'EARLY STOPPING AS THE MODEL NOT DOING WELL\n' + TextColor.END)
                return transducer_model, model_optimizer, stats

    if rank == 0:
        sys.stderr.write(TextColor.PURPLE + 'Finished training\n' + TextColor.END)

    return transducer_model, model_optimizer, stats


def cleanup():
    dist.destroy_process_group()


def setup(rank, device_ids, args):
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'

    # initialize the process group
    dist.init_process_group("gloo", rank=rank, world_size=len(device_ids))

    train_file, test_file, batch_size, epochs, gpu_mode, num_workers, retrain_model, \
    retrain_model_path, gru_layers, hidden_size, learning_rate, weight_decay, model_dir, stats_dir, total_callers, \
    train_mode = args

    # issue with semaphore lock: https://github.com/pytorch/pytorch/issues/2517
    # mp.set_start_method('spawn')

    # Explicitly setting seed to make sure that models created in two processes
    # start from same random weights and biases. https://github.com/pytorch/pytorch/issues/2517
    torch.manual_seed(42)
    train(train_file, test_file, batch_size, epochs, gpu_mode, num_workers, retrain_model, retrain_model_path,
          gru_layers, hidden_size, learning_rate, weight_decay, model_dir, stats_dir, train_mode,
          total_callers, rank, device_ids[rank])
    cleanup()


def train_distributed(train_file, test_file, batch_size, epochs, gpu_mode, num_workers, retrain_model,
                      retrain_model_path, gru_layers, hidden_size, learning_rate, weight_decay, model_dir,
                      stats_dir, device_ids, total_callers, train_mode):

    args = (train_file, test_file, batch_size, epochs, gpu_mode, num_workers, retrain_model,
            retrain_model_path, gru_layers, hidden_size, learning_rate, weight_decay, model_dir,
            stats_dir, total_callers, train_mode)
    mp.spawn(setup,
             args=(device_ids, args),
             nprocs=len(device_ids),
             join=True)
